from typing import Optional, Dict
import scanpy as sc
import networkx as nx
import numpy as np
import pandas as pd
import anndata as ad

import networkx as nx 

def filter_low_confidence_edges(adata: sc.AnnData,
                                edge_threshold: float,
                                flowsig_network_key: str = 'flowsig_network',
                                adjacency_key: str = 'adjacency',
                                filtered_key: str = 'filtered'):
    """
    Validate the learned CPDAG from UT-IGSP by checking edges against the assumed
    biological flow model, inflow signal -> gene expression module -> outflow signal.
    As the CPDAG contains directed arcs and undirected edges, we remove directed arcs
    that 

    Parameters
    ----------
    adata
        The annotated dataframe (typically from Scanpy) of the single-cell data.
        You need to have run FlowSig before running this step.

    edge_threshold
        The relative frequency of bootstrap edge frequency above which we keep edges.
        For directed arcs, we consider single edge frequencies. For undirected edges,
        we consider total edge weight.

    flowsig_network_key 
        The label in adata.uns where all of the flowsig output is stored, including the learned
        adjacency corresponding to the CPDAG (markov equivalence class), the flow variables used
        for inference, as well as their "flow types", which can be either inflow (ing), module
        (TFs or factors), or outflow (ing) signals.

    adjacency_key
        String key that specifies the adjacency for the learned network is stored in adata.uns[flowsig_network_key].

    adjacency_filtered_key
        String key that specifies where the validated network will be stored.

    Returns
    -------
    adjacency_filtered
        Matrix that encodes the CPDAG containing high-confidence directed arcs and 
        undirected arcs.

    """

    # Get the adjacency
    import graphical_models as gpm
    adjacency = adata.uns[flowsig_network_key]['network'][adjacency_key]
    flow_vars = adata.uns[flowsig_network_key]['flow_var_info'].index.tolist()

    cpdag = gpm.PDAG.from_amat(adjacency)

    adjacency_filtered = np.zeros(adjacency.shape)
    
    # First, let us calculate the total edge weights
    total_edge_weights = {}

    nonzero_rows, nonzero_cols = adjacency.nonzero()

    for i in range(len(nonzero_rows)):

        row_ind = nonzero_rows[i]
        col_ind = nonzero_cols[i]

        node_1 = flow_vars[row_ind]
        node_2 = flow_vars[col_ind]

        edge = (node_1, node_2)            

        # We either haven't recorded the edge, or we've taken the reverse edge previously
        if (edge[1], edge[0]) in total_edge_weights:
                total_edge_weights[(edge[1], edge[0])] += adjacency[row_ind, col_ind]
        else:
            total_edge_weights[edge] = adjacency[row_ind, col_ind]

    for arc in cpdag.arcs:

        node_1 = flow_vars[tuple(arc)[0]]
        node_2 = flow_vars[tuple(arc)[1]]

        row_ind = flow_vars.index(node_1)
        col_ind = flow_vars.index(node_2)

        edge_weight =  adjacency[row_ind, col_ind]

        # Need to account for both (node1, node2) and (node1, node1) as 
        # adjacency encodes directed network
        if edge_weight >= edge_threshold: 
            adjacency_filtered[row_ind, col_ind] = edge_weight

    for edge in cpdag.edges:

        node_1 = flow_vars[tuple(edge)[0]]
        node_2 = flow_vars[tuple(edge)[1]]

        # For directed arcs, we simply consider the total edge weights
        total_edge_weight = 0.0

        if (node_1, node_2) in total_edge_weights:

            total_edge_weight = total_edge_weights[(node_1, node_2)]

        else:

            total_edge_weight = total_edge_weights[(node_2, node_1)]

        # Need to account for both (node1, node2) and (node2, node1) as 
        # adjacency encodes directed network
        if total_edge_weight >= edge_threshold: 

            adjacency_filtered[tuple(edge)[0], tuple(edge)[1]] = adjacency[tuple(edge)[0], tuple(edge)[1]]
            adjacency_filtered[tuple(edge)[1], tuple(edge)[0]] = adjacency[tuple(edge)[1], tuple(edge)[0]]
            
    print(adjacency_filtered)

    # Save the "validated" adjacency
    filtered_adjacency_key = adjacency_key + '_' + filtered_key
    adata.uns[flowsig_network_key]['network'][filtered_adjacency_key] = adjacency_filtered

def apply_biological_flow(adata: sc.AnnData,
                        flowsig_network_key: str = 'flowsig_network',
                        adjacency_key: str = 'adjacency',
                        validated_key: str = 'validated'):
    """
    Validate the learned CPDAG from UT-IGSP by checking edges against the assumed
    biological flow model, inflow signal -> gene expression module -> outflow signal.
    As the CPDAG contains directed arcs and undirected edges, we remove directed arcs
    that do not follow these edge relations. For undirected edges that represent one of
    inflow -- gene expression module, gene expression module -- gene expression module,
    and gene expression module -- outflow, we orient them so that they make "biological
    sense".

    Parameters
    ----------
    adata
        The annotated dataframe (typically from Scanpy) of the single-cell data.
        You need to have run FlowSig before running this step.

    flowsig_network_key 
        The label in adata.uns where all of the flowsig output is stored, including the learned
        adjacency corresponding to the CPDAG (markov equivalence class), the flow variables used
        for inference, as well as their "flow types", which can be either inflow (ing), module (TFs or factors),
        or outflow (ing) signals.

    adjacency_key
        String key that specifies the adjacency for the learned network is stored in adata.uns[flowsig_network_key].

    adjacency_validated_key
        String key that specifies where the validated network will be stored.

    Returns
    -------
    adjacency_validated
        Matrix that encodes the CPDAG containing "biologically realistic" inferred flows, from
        inflow variables, to gene expression module variables, to outflow variables.

    """

    # Get the adjacency
    import graphical_models as gpm
    adjacency = adata.uns[flowsig_network_key]['network'][adjacency_key]
    flow_vars = adata.uns[flowsig_network_key]['flow_var_info'].index.tolist()
    #make sure the index is unique
    adata.uns[flowsig_network_key]['flow_var_info'].index=ad.utils.make_index_unique(adata.uns[flowsig_network_key]['flow_var_info'].index)
    flow_var_info = adata.uns[flowsig_network_key]['flow_var_info']

    cpdag = gpm.PDAG.from_amat(adjacency)
    adjacency_validated = np.zeros(adjacency.shape)
    
    for arc in cpdag.arcs:

        node_1 = flow_vars[tuple(arc)[0]]
        node_2 = flow_vars[tuple(arc)[1]]

        # Classify node types
        node_1_type = flow_var_info.loc[node_1]['Type']
        node_2_type = flow_var_info.loc[node_2]['Type']

        # Now we decide whether or not to add the  edges
        add_edge = False

        # Define the edge because we may need to reverse it
        edge = (node_1, node_2)

        # If there's a link from received morphogen to a TF
        if ( (node_1_type == 'inflow')&(node_2_type == 'module') ):

            add_edge = True

        if ( (node_1_type == 'module')&(node_2_type == 'outflow') ):

            add_edge = True

        if ((node_1_type == 'module')&(node_2_type == 'module')):

            add_edge = True

        if add_edge:

            row_ind = tuple(arc)[0]
            col_ind = tuple(arc)[1]

            adjacency_validated[row_ind, col_ind] = adjacency[row_ind, col_ind]

    for edge in cpdag.edges:

        node_1 = flow_vars[tuple(edge)[0]]
        node_2 = flow_vars[tuple(edge)[1]]

        # Classify node types
        node_1_type = flow_var_info.loc[node_1]['Type']
        node_2_type = flow_var_info.loc[node_2]['Type']

        # Define the edge because we may need to reverse it
        add_edge = False

        # If there's a link from received morphogen to a TF
        if ( (node_1_type == 'inflow')&(node_2_type == 'module') ):

            add_edge = True

        # If there's a link from received morphogen to a TF
        if ( (node_1_type == 'module')&(node_2_type == 'inflow') ):

            add_edge = True

        if ( (node_1_type == 'module')&(node_2_type == 'outflow') ):

            add_edge = True

        if ( (node_1_type == 'outflow')&(node_2_type == 'module') ):

            add_edge = True

        if ((node_1_type == 'module')&(node_2_type == 'module')):

            add_edge = True

        if add_edge:

            row_ind = tuple(edge)[0]
            col_ind = tuple(edge)[1]

            adjacency_validated[row_ind, col_ind] = adjacency[row_ind, col_ind]

            adjacency_validated[col_ind, row_ind] = adjacency[col_ind, row_ind]
    
    # Save the "validated" adjacency
    print(adjacency_validated)

    validated_adjacency_key = adjacency_key + '_' + validated_key
    adata.uns[flowsig_network_key]['network'][validated_adjacency_key] = adjacency_validated

def construct_intercellular_flow_network(adata: sc.AnnData,
                        flowsig_network_key: str = 'flowsig_network',
                        adjacency_key: str = 'adjacency'):
    import graphical_models as gpm
    flow_vars = adata.uns[flowsig_network_key]['flow_var_info'].index.tolist()
    flow_var_info = adata.uns[flowsig_network_key]['flow_var_info']

    flow_adjacency = adata.uns[flowsig_network_key]['network'][adjacency_key]

    nonzero_rows, nonzero_cols = flow_adjacency.nonzero()

    total_edge_weights = {}

    for i in range(len(nonzero_rows)):
        row_ind = nonzero_rows[i]
        col_ind = nonzero_cols[i]

        node_1 = flow_vars[row_ind]
        node_2 = flow_vars[col_ind]

        edge = (node_1, node_2)            

        if ( (edge not in total_edge_weights)&((edge[1], edge[0]) not in total_edge_weights) ):
            total_edge_weights[edge] = flow_adjacency[row_ind, col_ind]
        else:
            if (edge[1], edge[0]) in total_edge_weights:
                total_edge_weights[(edge[1], edge[0])] += flow_adjacency[row_ind, col_ind]
    
    flow_network = nx.DiGraph()

    # Now let's consturct the graph from the CPDAG
    cpdag =  gpm.PDAG.from_amat(flow_adjacency)

    # Add the directed edges (arcs) first
    for arc in cpdag.arcs:

        node_1 = flow_vars[tuple(arc)[0]]
        node_2 = flow_vars[tuple(arc)[1]]

        # Classify node types
        node_1_type = flow_var_info.loc[node_1]['Type']
        node_2_type = flow_var_info.loc[node_2]['Type']

        # Now we decide whether or not to add the damn edges
        add_edge = False

        # Define the edge because we may need to reverse it
        edge = (node_1, node_2)

        # If there's a link from the expressed morphogen to the received morphogen FOR the same morphogen
        if ( (node_1_type == 'inflow')&(node_2_type == 'module') ):
            add_edge = True

        # If there's a link from received morphogen to a TF
        if ( (node_1_type == 'module')&(node_2_type == 'outflow') ):

            add_edge = True

        if ( (node_1_type == 'module')&(node_2_type == 'module') ):

            add_edge = True

        if add_edge:

            # Get the total edge weight
            total_edge_weight = 0.0

            if edge in total_edge_weights:

                total_edge_weight = total_edge_weights[edge]

            else:

                total_edge_weight = total_edge_weights[(edge[1], edge[0])]

            edge_weight = flow_adjacency[tuple(arc)[0], tuple(arc)[1]]

            flow_network.add_edge(*edge)
            flow_network.edges[edge[0], edge[1]]['weight'] = edge_weight / total_edge_weight
            flow_network.nodes[edge[0]]['type'] = node_1_type
            flow_network.nodes[edge[1]]['type'] = node_2_type

    for edge in cpdag.edges:

        node_1 = flow_vars[tuple(edge)[0]]
        node_2 = flow_vars[tuple(edge)[1]]

        # Classify node types
        node_1_type = flow_var_info.loc[node_1]['Type']
        node_2_type = flow_var_info.loc[node_2]['Type']

        # Define the edge because we may need to reverse it
        undirected_edge = (node_1, node_2)

        add_edge = False
        # If there's a link from the expressed morphogen to the received morphogen FOR the same morphogen
        if ( (node_1_type == 'inflow')&(node_2_type == 'module') ):

            add_edge = True

        if ( (node_1_type == 'module')&(node_2_type == 'inflow') ):

            add_edge = True
            undirected_edge = (node_2, node_1)

        if ( (node_1_type == 'module')&(node_2_type == 'outflow') ):

            add_edge = True

        if ( (node_1_type == 'outflow')&(node_2_type == 'module') ):

            add_edge = True
            undirected_edge = (node_2, node_1)

        if ((node_1_type == 'module')&(node_2_type == 'module')):

            add_edge = True

        if add_edge:

            # Get the total edge weight
            total_edge_weight = 0.0

            if undirected_edge in total_edge_weights:

                total_edge_weight = total_edge_weights[undirected_edge]

            else:

                total_edge_weight = total_edge_weights[(undirected_edge[1], undirected_edge[0])]

            flow_network.add_edge(*undirected_edge)
            flow_network.edges[undirected_edge[0], undirected_edge[1]]['weight'] = min(total_edge_weight, 1.0)
            flow_network.nodes[undirected_edge[0]]['type'] = node_1_type
            flow_network.nodes[undirected_edge[1]]['type'] = node_2_type

            # Add the other way if we have modules
            if ((node_1_type == 'module')&(node_2_type == 'module')):

                flow_network.add_edge(undirected_edge[1], undirected_edge[0])
                flow_network.edges[undirected_edge[1], undirected_edge[0]]['weight'] = min(total_edge_weight, 1.0)
                flow_network.nodes[undirected_edge[0]]['type'] = node_2_type
                flow_network.nodes[undirected_edge[1]]['type'] = node_1_type

    return flow_network


#聚类
from sklearn.mixture import GaussianMixture
import scanpy as sc
import pandas as pd
import numpy as np
import anndata
from .._settings import add_reference

mira_install=False
def global_imports(modulename,shortname = None, asfunction = False):
    if shortname is None: 
        shortname = modulename
    if asfunction is False:
        globals()[shortname] = __import__(modulename)
    else:        
        globals()[shortname] = __import__(modulename)

#初始化聚类位置，这个很重要
def get_initial_means(X, n_components,init_params, r):
    # Run a GaussianMixture with max_iter=0 to output the initialization means
    gmm = GaussianMixture(
        n_components=n_components, init_params=init_params, tol=1e-9, max_iter=0, random_state=r
    ).fit(X)
    return gmm.means_


def mclust_py(adata,  n_components=None,use_rep:str='X_pca',
              modelNames='EEE',  random_seed=2020):
    """Clustering using Gaussian Mixture Model (GMM), similar to mclust in R."""

    if n_components is None:
        print('You need to input the `n_components` when methods is `GMM`')
        return
    print(f"""running GaussianMixture clustering""")
    # Extract the data to be clustered
    data = adata.obsm[use_rep]
    
    import numpy as np
    np.random.seed(random_seed)
    
    # Extract the data to be clustered
    data = adata.obsm[use_rep]
    
    # Map modelNames to scikit-learn covariance_type
    covariance_type_map = {
        'EEE': 'spherical',  # Equal volume, shape, and orientation (spherical)
        'VVV': 'full',       # Variable volume, shape, and orientation
        'EEV': 'tied',       # Equal volume and shape, variable orientation (tied)
        'VVI': 'diag',       # Variable volume and shape, equal orientation (diag)
        # Add more mappings as needed
    }
    
    covariance_type = covariance_type_map.get(modelNames, 'full')
    
    # Initialize and fit the Gaussian Mixture Model
    gmm = GaussianMixture(n_components=n_components, covariance_type=covariance_type, random_state=random_seed)
    gmm.fit(data)
    
    # Get the cluster labels
    mclust_res = gmm.predict(data)
    
    # Add the cluster labels to adata.obs
    adata.obs['mclust'] = mclust_res
    adata.obs['mclust'] = adata.obs['mclust'].astype('int')
    adata.obs['mclust'] = adata.obs['mclust'].astype('category')
    adata.obs['gmm_cluster'] = adata.obs['mclust']
    
    return adata


def cluster(adata:anndata.AnnData,method:str='leiden',
            use_rep:str='X_pca',random_state:int=1024,
            n_components=None, **kwargs):
    
    if method=='leiden':
        sc.tl.leiden(adata,**kwargs)
        add_reference(adata,'leiden','clustering with Leiden')
    elif method=='louvain':
        sc.tl.louvain(adata,**kwargs)
        add_reference(adata,'louvain','clustering with Louvain')
    elif method=='GMM':
        mclust_py(adata, n_components=n_components,use_rep=use_rep,random_seed=random_state,**kwargs)
        print(f"""finished: found {n_components} clusters and added
    'mclust', the cluster labels (adata.obs, categorical)""")
        add_reference(adata,'GMM','clustering with Gaussian Mixture Model')
    elif method=='mclust':
        mclust_py(adata, n_components=n_components,use_rep=use_rep,
                  random_seed=random_state,**kwargs)
        print(f"""finished: found {n_components} clusters and added
    'mclust', the cluster labels (adata.obs, categorical)""")
        add_reference(adata,'mclust','clustering with Gaussian Mixture Model')
    elif method=='schist':
        try:
            import schist
        except ImportError:
            raise ImportError(
                'Please install the schist using conda `conda install -c conda-forge schist` \nor `pip install git+https://github.com/dawe/schist.git`'
            )
        schist.inference.nested_model(adata, **kwargs)
        add_reference(adata,'schist','clustering with schist')
    elif method=='mclust_R':
        np.random.seed(random_state)
        import rpy2.robjects as robjects
        robjects.r.library("mclust")

        import rpy2.robjects.numpy2ri
        rpy2.robjects.numpy2ri.activate()
        r_random_seed = robjects.r['set.seed']
        r_random_seed(random_state)
        rmclust = robjects.r['Mclust']

        res = rmclust(rpy2.robjects.numpy2ri.numpy2rpy(adata.obsm[use_rep]), n_components, 'EEE')
        mclust_res = np.array(res[-2])

        adata.obs[method] = mclust_res
        adata.obs[method] = adata.obs[method].astype('int')
        adata.obs[method] = adata.obs[method].astype('category')
        print(f"""finished: found {n_components} clusters and added
    'mclust', the cluster labels (adata.obs, categorical)""")
        add_reference(adata,'mclust','clustering with Gaussian Mixture Model')

      
def refine_label(adata, use_rep='spatial',radius=50, key='label'):
    """
    Optimize the label by majority voting in the neighborhood.

    Args:
        adata: an Anndata object, after normalization.
        radius: the radius of the neighborhood.
        key: the key in `.obs` that corresponds to the cluster labels.
    """
    from scipy.spatial import distance
    from tqdm import tqdm
    n_neigh = radius
    new_type = []
    old_type = adata.obs[key].values

    # calculate distance
    position = adata.obsm[use_rep]
    dist_matrix = distance.cdist(position, position, metric="euclidean")

    n_cell = dist_matrix.shape[0]

    for i in tqdm(range(n_cell)):
        vec = dist_matrix[i, :]
        index = vec.argsort()
        neigh_type = []
        for j in range(1, n_neigh + 1):
            neigh_type.append(old_type[index[j]])
        max_type = max(neigh_type, key=neigh_type.count)
        new_type.append(max_type)

    new_type = [str(i) for i in list(new_type)]
    # adata.obs['label_refined'] = np.array(new_type)

    return new_type
        
def filtered(adata:anndata.AnnData,
             cluster_key:str,
             cluster_minsize:int=10):
    new_num=adata.obs[cluster_key].value_counts()[adata.obs[cluster_key].value_counts()<cluster_minsize].shape[0]
    adata.obs['gmm_cluster']=adata.obs['gmm_cluster'].astype(str)
    adata.obs.loc[adata.obs[cluster_key].isin(adata.obs[cluster_key].value_counts()[adata.obs[cluster_key].value_counts()<cluster_minsize].index.tolist()),cluster_key]='-1'
    adata.obs[cluster_key]=adata.obs[cluster_key].astype('category')
    adata.obs['gmm_cluster'].cat.categories=pd.Index(list(range(len(adata.obs['gmm_cluster'].cat.categories))))
    print(f"""filtered {new_num} clusters and changed the cluster labels to '-1'(adata.obs, categorical)""")


class LDA_topic(object):

    def __init__(self,adata,feature_type='expression',
                  highly_variable_key='highly_variable_features',
                 layers='counts',batch_key=None,learning_rate=1e-3,ondisk=False):
        global mira_install
        try:
            import mira
            mira_install=True
            print('mira have been install version:',mira.__version__)
        except ImportError:
            raise ImportError(
                """Please install the mira: `conda install -c bioconda mira-multiome` or 
                `pip install mira-multiome`.'"""
            )
        if mira_install==True:
            global_imports("mira")
        self.adata=adata
        self.model = mira.topics.make_model(
            adata.n_obs, adata.n_vars, # helps MIRA choose reasonable values for some hyperparameters which are not tuned.
            feature_type = feature_type,
            highly_variable_key=highly_variable_key,
            counts_layer=layers,
            categorical_covariates=batch_key
        )
        self.ondisk=ondisk
        if self.ondisk!=True:
        
            self.model.get_learning_rate_bounds(adata)
            self.model.set_learning_rates(learning_rate, 0.25) # for larger datasets, the default of 1e-3, 0.1 usually works well.
        else:
            train, test = self.model.train_test_split(adata)
            import os 
            os.mkdir('topic_train')
            os.mkdir('topic_test')
            self.model.write_ondisk_dataset(train, dirname='topic_train')
            self.model.write_ondisk_dataset(test, dirname='topic_test')
            self.model.get_learning_rate_bounds('topic_train')
            self.model.set_learning_rates(learning_rate, 0.25) # for larger datasets, the default of 1e-3, 0.1 usually works well.

        self.model.plot_learning_rate_bounds(figsize=(6,3))
        
    def plot_topic_contributions(self,num_topics=6):
        NUM_TOPICS = num_topics
        if self.ondisk!=True:
            topic_contributions = mira.topics.gradient_tune(self.model, self.adata)
        else:
            topic_contributions = mira.topics.gradient_tune(self.model, 'topic_train')
        mira.pl.plot_topic_contributions(topic_contributions, NUM_TOPICS)
        
    def predicted(self,num_topics=6):
        print(f"""running LDA topic predicted""")
        self.model = self.model.set_params(num_topics = num_topics).fit(self.adata)
        self.model.predict(self.adata)
        # 找到每行中最大值所在的列（topic）
        df=self.adata.obs[self.model.topic_cols].copy()
        max_topic = df.idxmax(axis=1)
        # 将结果添加到DataFrame中
        self.adata.obs['LDA_cluster'] = max_topic
        print(f"""finished: found {num_topics} clusters and added
    'LDA_cluster', the cluster labels (adata.obs, categorical)""")
        
    def get_results_rfc(self,adata,use_rep='X_pca',LDA_threshold=0.5,num_topics=6):
        import pandas as pd
        print(f"""running LDA topic predicted""")
        #self.model = self.model.set_params(num_topics = num_topics).fit(self.adata)
        #self.model.predict(self.adata)
        # 找到每行中最大值所在的列（topic）
        #df=self.adata.obs[self.model.topic_cols].copy()
        
        import numpy as np
        import pandas as pd
        import matplotlib.pyplot as plt
        from sklearn.tree import DecisionTreeClassifier
        from sklearn.ensemble import RandomForestClassifier
        from sklearn.model_selection import train_test_split

        new_array = []
        class_array = []
        for i in range(0, num_topics):
            data = adata[adata.obs[f'topic_{i}'] > LDA_threshold].obsm[use_rep].toarray()
            new_array.append(data)
            class_array.append(np.full(data.shape[0], i))

        new_array = np.concatenate(new_array, axis=0)
        class_array = np.concatenate(class_array)

        Xtrain, Xtest, Ytrain, Ytest = train_test_split(new_array,class_array,test_size=0.3)
        clf = DecisionTreeClassifier(random_state=0)
        rfc = RandomForestClassifier(random_state=0)
        clf = clf.fit(Xtrain,Ytrain)
        rfc = rfc.fit(Xtrain,Ytrain)
        #查看模型效果
        score_c = clf.score(Xtest,Ytest)
        score_r = rfc.score(Xtest,Ytest)
        #打印最后结果
        print("Single Tree:",score_c)
        print("Random Forest:",score_r)

        adata.obs['LDA_cluster_rfc']=[str(i) for i in rfc.predict(adata.obsm[use_rep])]
        adata.obs['LDA_cluster_clf']=[str(i) for i in clf.predict(adata.obsm[use_rep])]
        print('LDA_cluster_rfc is added to adata.obs')
        print('LDA_cluster_clf is added to adata.obs')


import numpy as np
import pandas as pd
from tqdm import tqdm
import scipy.sparse as sp

from .STAGATE import STAGATE
from .utils import Transfer_pytorch_Data

import torch
import torch.backends.cudnn as cudnn
cudnn.deterministic = True
cudnn.benchmark = True
import torch.nn.functional as F


def train_STAGATE(adata, hidden_dims=[512, 30], n_epochs=1000, lr=0.001, key_added='STAGATE',
                gradient_clipping=5.,  weight_decay=0.0001, verbose=True, 
                random_seed=0, save_loss=False, save_reconstrction=False, 
                device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')):
    """\
    Training graph attention auto-encoder.

    Parameters
    ----------
    adata
        AnnData object of scanpy package.
    hidden_dims
        The dimension of the encoder.
    n_epochs
        Number of total epochs in training.
    lr
        Learning rate for AdamOptimizer.
    key_added
        The latent embeddings are saved in adata.obsm[key_added].
    gradient_clipping
        Gradient Clipping.
    weight_decay
        Weight decay for AdamOptimizer.
    save_loss
        If True, the training loss is saved in adata.uns['STAGATE_loss'].
    save_reconstrction
        If True, the reconstructed expression profiles are saved in adata.layers['STAGATE_ReX'].
    device
        See torch.device.

    Returns
    -------
    AnnData
    """

    # seed_everything()
    seed=random_seed
    import random
    random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    np.random.seed(seed)

    adata.X = sp.csr_matrix(adata.X)
    
    if 'highly_variable' in adata.var.columns:
        adata_Vars =  adata[:, adata.var['highly_variable']]
    else:
        adata_Vars = adata

    if verbose:
        print('Size of Input: ', adata_Vars.shape)
    if 'Spatial_Net' not in adata.uns.keys():
        raise ValueError("Spatial_Net is not existed! Run Cal_Spatial_Net first!")

    data = Transfer_pytorch_Data(adata_Vars)

    model = STAGATE(hidden_dims = [data.x.shape[1]] + hidden_dims).to(device)
    data = data.to(device)

    optimizer = torch.optim.Adam(model.parameters(), lr=lr, weight_decay=weight_decay)

    #loss_list = []
    for epoch in tqdm(range(1, n_epochs+1)):
        model.train()
        optimizer.zero_grad()
        z, out = model(data.x, data.edge_index)
        loss = F.mse_loss(data.x, out) #F.nll_loss(out[data.train_mask], data.y[data.train_mask])
        #loss_list.append(loss)
        loss.backward()
        torch.nn.utils.clip_grad_norm_(model.parameters(), gradient_clipping)
        optimizer.step()
    
    model.eval()
    z, out = model(data.x, data.edge_index)
    
    STAGATE_rep = z.to('cpu').detach().numpy()
    adata.obsm[key_added] = STAGATE_rep

    if save_loss:
        adata.uns['STAGATE_loss'] = loss
    if save_reconstrction:
        ReX = out.to('cpu').detach().numpy()
        ReX[ReX<0] = 0
        adata.layers['STAGATE_ReX'] = ReX

    return adata



r"""
Vis multi dataset and their connection
"""
import random
import math
from typing import List, Mapping, Optional, Union

import scanpy as sc
import numpy as np
import pandas as pd
from anndata import AnnData

import matplotlib.pyplot as plt
import plotly.graph_objects as go

from .color import *


def get_color(n=1, cmap:str='scanpy', seed:int=0):
    r"""
    Get color 
    
    Parameters
    ---------
    n
        number of colors you want
    cmap
        color map (use same with scanpy)
    seed
        random seed to duplicate
    """
    if cmap == 'scanpy' and n <= 10:
        step = 10 // n
        return vega_10_scanpy[::step][:n]
    elif cmap == 'scanpy' and n <= 20 :
        step = 20 // n
        return vega_20_scanpy[::step][:n]
    elif cmap == 'scanpy' and n <= 28:
        step = 28 // n
        return zeileis_28[::step][:n]
    elif cmap == 'scanpy' and n <= 102:
        step = 102 // n
        return godsnot_102[::step][:n]
    else:
        print('WARNING: Using random color')
        random.seed(seed)
        if n==1:
            return "#"+''.join([random.choice('0123456789ABCDEF') for j in range(6)])
        elif n>1 :
            return ["#"+''.join([random.choice('0123456789ABCDEF') for j in range(6)]) for i in range(n)]


class build_3D():
    r"""
    Build 3D pics/models from multi-datasets
    
    Parameters
    ---------
    datasets
        list adata of in order
    mappings
        list of SLAT matching results
    spatial_key
        obsm key of spatial info
    anno_key
        obs key of cell annotation such as celltype
    subsample_size
        subsample size of matches
    scale_coordinate
        scale the coordinate from different slides
    """
    def __init__(self,adatas:List[AnnData],
               mappings:List[np.ndarray],
               spatial_key:Optional[str]='spatial',
               anno_key:Optional[str]='annotation',
               subsample_size:Optional[int]=200,
               scale_coordinate:Optional[bool]=True,
        ) -> None:
        assert len(mappings) == len(adatas) - 1
        
        self.mappings=mappings
        self.loc_list = []
        self.anno_list = []
        for adata in adatas:
            loc = adata.obsm[spatial_key].copy()
            if scale_coordinate:
                for i in range(2):
                    loc[:,i] = (loc[:,i]-np.min(loc[:,i]))/(np.max(loc[:,i])-np.min(loc[:,i]))
            anno = adata.obs[anno_key]
            self.loc_list.append(loc)
            self.anno_list.append(anno)
            
        self.adatas = adatas
        self.anno_key=anno_key
        self.celltypes = set(pd.concat(self.anno_list))
        self.subsample_size = subsample_size
            
    def draw_3D(self,
                size: Optional[List[int]]=[10,10],
                point_size: Optional[List[int]]=[0.5,0.5],
                point_alpha: Optional[float]=0.6,
                line_width: Optional[float]=0.6,
                line_color: Optional[str]='#4169E1',
                line_alpha: Optional[float]=0.8,
                hide_axis: Optional[bool]=False,
                height: Optional[float]=1.0,
                height_scale: Optional[float]=1.0,
        ) -> None:
        r"""
        Draw 3D picture of two layers
        
        Parameters:
        ----------
        size
            plt figure size (width, height)
        point_size
            point size of each layer
        point_alpha
            point alpha of each layer
        line_width
            pair line width
        line_color
            pair line color
        line_alpha
            pair line alpha
        hide_axis
            if hide axis
        height
            height of one layer
        """
        fig = plt.figure(figsize=(size[0],size[1]))
        ax = fig.add_subplot(111, projection='3d')
        ax.set_box_aspect([1, 1, height_scale * len(self.mappings)])
        # color by different cell types
        
        

        color = get_color(len(self.celltypes))
        c_map = {}
        for i, celltype in enumerate(self.celltypes):
            c_map[celltype] = color[i]
        for j, mapping in enumerate(self.mappings):
            print(f"Mapping {j}th layer ")
            # plot cells
            for i, (layer, anno,ad) in enumerate(zip(self.loc_list[j:j+2], self.anno_list[j:j+2],self.adatas[j:j+2])):
                if i==0 and 0<j<len(self.mappings)-1:
                    continue
                
                ad.obs[self.anno_key]=ad.obs[self.anno_key].astype('category')
                if '{}_colors'.format(self.anno_key) in ad.uns.keys():
                    c_map=dict(zip(ad.obs[self.anno_key].cat.categories.tolist(),
                                    ad.uns['{}_colors'.format(self.anno_key)]))
                else:
                    if len(ad.obs[self.anno_key].cat.categories)>28:
                        c_map=dict(zip(ad.obs[self.anno_key].cat.categories,sc.pl.palettes.default_102))
                    else:
                        c_map=dict(zip(ad.obs[self.anno_key].cat.categories,sc.pl.palettes.zeileis_28))

                for cell_type in ad.obs[self.anno_key].cat.categories:
                    slice = layer[anno == cell_type,:]
                    xs = slice[:,0]
                    ys = slice[:,1]
                    zs = height*(j+i)
                    ax.scatter(xs, ys, zs, s=point_size[i], c=c_map[cell_type], alpha=point_alpha)
            # plot mapping line
            mapping = mapping[:,np.random.choice(mapping.shape[1], self.subsample_size, replace=False)].copy()
            for k in range(mapping.shape[1]):
                cell1_index = mapping[:,k][0]  # query
                cell0_index = mapping[:,k][1]  # ref
                cell0_coord = self.loc_list[j][cell0_index,:]
                cell1_coord = self.loc_list[j+1][cell1_index,:]
                coord = np.row_stack((cell0_coord,cell1_coord))
                ax.plot(coord[:,0], coord[:,1], [height*j, height*(j+1)], color=line_color,
                        linestyle="dashed", linewidth=line_width, alpha=line_alpha)

        if hide_axis:
            plt.axis('off')
        return ax
        #plt.show()


class match_3D_multi():
    r"""
    Plot the mapping result between 2 datasets
    
    Parameters
    ---------
    dataset_A
        pandas dataframe which contain ['index','x','y'], reference dataset
    dataset_B
        pandas dataframe which contain ['index','x','y'], target dataset
    matching
        matching results
    meta
        dataframe colname of meta, such as celltype
    expr
        dataframe colname of gene expr
    subsample_size
        subsample size of matches
    reliability
        match score (cosine similarity score)
    scale_coordinate
        if scale coordinate via (:math:`data - np.min(data)) / (np.max(data) - np.min(data))`)
    rotate
        how to rotate the slides (force scale_coordinate), such as ['x','y'], means dataset0 rotate on x axes
        and dataset1 rotate on y axes
    change_xy
        exchange x and y on dataset_B
    subset
        index of query cells to be plotted

    Note
    ----------
    dataset_A and dataset_B can in different length
        
    """
    def __init__(self,dataset_A:pd.DataFrame,
                dataset_B:pd.DataFrame,
                matching:np.ndarray,
                meta:Optional[str]=None,
                expr:Optional[str]=None,
                subsample_size:Optional[int]=300,
                reliability:Optional[np.ndarray]=None,
                scale_coordinate:Optional[bool]=True,
                rotate:Optional[List[str]]=None,
                exchange_xy:Optional[bool]=False,
                subset: Optional[List[int]]=None
        ) -> None:
        self.dataset_A = dataset_A.copy()
        self.dataset_B = dataset_B.copy()
        self.meta = meta
        self.matching= matching
        self.conf = reliability
        self.subset = subset # index of query cells to be plotted
        scale_coordinate = True if rotate != None else scale_coordinate
        
        assert all(item in dataset_A.columns.values for item in ['index','x','y'])
        assert all(item in dataset_B.columns.values for item in ['index','x','y'])
        
        if meta:
            set1 = list(set(self.dataset_A[meta]))
            set2 = list(set(self.dataset_B[meta]))
            self.celltypes = set1 + [x for x in set2 if x not in set1]
            self.celltypes.sort() # make sure celltypes are in the same order
            overlap = [x for x in set2 if x in set1]
            print(f"dataset1: {len(set1)} cell types; dataset2: {len(set2)} cell types; \n\
                    Total :{len(self.celltypes)} celltypes; Overlap: {len(overlap)} cell types \n\
                    Not overlap :[{[y for y in (set1+set2) if y not in overlap]}]"
                    )
        self.expr = expr if expr else False
            
        if scale_coordinate:
            for i, dataset in enumerate([self.dataset_A, self.dataset_B]):
                for axis in ['x','y']:
                    dataset[axis] = (dataset[axis] - np.min(dataset[axis])) / (np.max(dataset[axis])- np.min(dataset[axis]))
                    if rotate == None:
                        pass
                    elif axis in rotate[i]:
                        dataset[axis] = 1 - dataset[axis]
        if exchange_xy:
            self.dataset_B[['x','y']] = self.dataset_B[['y','x']]

        if not subset is None:
            matching = matching[:,subset]
        if matching.shape[1] > subsample_size and subsample_size > 0:
            self.matching = matching[:,np.random.choice(matching.shape[1],subsample_size, replace=False)]
        else:
            subsample_size = matching.shape[1]
            self.matching = matching
        print(f'Subsampled {subsample_size} pairs from {matching.shape[1]}')
        
            
        self.datasets = [self.dataset_A, self.dataset_B]
    
    def draw_3D(self,
                size: Optional[List[int]]=[10,10],
                conf_cutoff: Optional[float]=0,
                point_size: Optional[List[int]]=[0.1,0.1],
                line_width: Optional[float]=0.3,
                line_color:Optional[str]='grey',
                line_alpha: Optional[float]=0.7,
                hide_axis: Optional[bool]=False,
                show_error: Optional[bool]=True,
                show_celltype: Optional[bool]=False,
                cmap: Optional[bool]='Reds',
                save:Optional[str]=None
        ) -> None:
        r"""
        Draw 3D picture of two datasets
        
        Parameters:
        ----------
        size
            plt figure size
        conf_cutoff
            confidence cutoff of mapping to be plotted
        point_size
            point size of every dataset
        line_width
            pair line width
        line_color
            pair line color
        line_alpha
            pair line alpha
        hide_axis
            if hide axis
        show_error
            if show error celltype mapping with different color
        cmap
            color map when vis expr
        save
            save file path
        """
        self.conf_cutoff = conf_cutoff
        show_error = show_error if self.meta else False
        fig = plt.figure(figsize=(size[0],size[1]))
        ax = fig.add_subplot(111, projection='3d')
        # color by meta
        if self.meta:
            color = get_color(len(self.celltypes))
            c_map = {}
            for i, celltype in enumerate(self.celltypes):
                c_map[celltype] = color[i]
            if self.expr:
                c_map = cmap
                # expr_concat = pd.concat(self.datasets)[self.expr].to_numpy()
                # norm = plt.Normalize(expr_concat.min(), expr_concat.max())
            for i, dataset in enumerate(self.datasets):
                if self.expr:
                    norm = plt.Normalize(dataset[self.expr].to_numpy().min(), dataset[self.expr].to_numpy().max())
                
                for cell_type in list(set(dataset[self.meta])):
                    slice = dataset[dataset[self.meta] == cell_type]
                    xs = slice['x']
                    ys = slice['y']
                    if 'color' in slice.columns:
                        cs=slice['color'].tolist()[0]
                    else:
                        cs = c_map[cell_type]
                    zs = i
                    if self.expr:
                        ax.scatter(xs, ys, zs, s=point_size[i], c=slice[self.expr], cmap=c_map, norm=norm)
                    else:
                        ax.scatter(xs, ys, zs, s=point_size[i], c=cs)
        # plot points without meta
        else:
            for i, dataset in enumerate(self.datasets):
                xs = dataset['x']
                ys = dataset['y']
                zs = i
                ax.scatter(xs,ys,zs,s=point_size[i])
        # plot line
        self.c_map = c_map
        self.draw_lines(ax, show_error, show_celltype, line_color, line_width, line_alpha)
        if hide_axis:
            plt.axis('off')
        if save != None:
            plt.savefig(save)
        plt.show()
        
    def draw_lines(self, ax, show_error, show_celltype, line_color, line_width=0.3, line_alpha=0.7) -> None:
        r"""
        Draw lines between paired cells in two datasets
        """
        for i in range(self.matching.shape[1]):
            if not self.conf is None and self.conf[i] < self.conf_cutoff:
                continue
            pair = self.matching[:,i]
            default_color = line_color
            if self.meta != None:
                celltype1 = self.dataset_A.loc[self.dataset_A['index']==pair[1], self.meta].astype(str).values[0]
                celltype2 = self.dataset_B.loc[self.dataset_B['index']==pair[0], self.meta].astype(str).values[0]
                if show_error:
                    if celltype1 == celltype2:
                        color = '#ade8f4' # blue
                    else:
                        color = '#ffafcc'  # red
                if show_celltype:
                    if celltype1 == celltype2:
                        color = self.c_map[celltype1]
                    else:
                        color = '#696969' # celltype1 error match color
            point0 = np.append(self.dataset_A[self.dataset_A['index']==pair[1]][['x','y']], 0)
            point1 = np.append(self.dataset_B[self.dataset_B['index']==pair[0]][['x','y']], 1)

            coord = np.row_stack((point0, point1))
            color = color if show_error or show_celltype else default_color
            ax.plot(coord[:,0], coord[:,1], coord[:,2], color=color, linestyle="dashed", linewidth=line_width, alpha=line_alpha)


class match_3D_multi_error(match_3D_multi):
    r"""
    Highlight the error mapping between datasets, child of class:`match_3D_multi()`
    
    Parameters
    ---------
    dataset_A
        pandas dataframe which contain ['index','x','y']
    dataset_B
        pandas dataframe which contain ['index','x','y']
    matching
        matching results
    mode
        which cell pairs to highlight
    highlight_color
        color to highlight the line 
    meta
        dataframe colname of meta, such as celltype
    expr
        dataframe colname of gene expr
    subsample_size
        subsample size of matches
    reliability
        if the match is reliable
    scale_coordinate
        if scale the coordinate via `data - np.min(data)) / (np.max(data) - np.min(data))` 
    rotate
        how to rotate the slides (force scale_coordinate)
    change_xy
        exchange x and y on dataset_B
    subset
        index of query cells to be plotted
        
    Note
    ----------
    dataset_A and dataset_B can in different length
        
    """
    def __init__(self,dataset_A: pd.DataFrame,
                 dataset_B: pd.DataFrame,
                 matching: np.ndarray,
                 mode: Optional[str]='high_true',
                 highlight_color: Optional[str]='red',
                 meta: Optional[str]=None,
                 expr: Optional[str]=None,
                 subsample_size: Optional[int]=300,
                 reliability: Optional[np.ndarray]=None,
                 scale_coordinate: Optional[bool]=False,
                 rotate: Optional[List[str]]=None,
                 exchange_xy: Optional[bool]=False,
                 subset: Optional[Union[np.ndarray,List[int]]]=None
        ) -> None:
        super(match_3D_multi_error, self).__init__(dataset_A,dataset_B,matching,meta,expr,subsample_size,reliability,
                                                   scale_coordinate,rotate,exchange_xy,subset)
        assert mode in ['high_true','low_true','high_false','low_false']
        self.mode = mode
        self.highlight_color = highlight_color
        
    def draw_lines(self, ax, show_error, show_celltype, default_color, line_width=0.3, line_alpha=0.7) -> None:
        for i in range(self.matching.shape[1]):
            pair = self.matching[:,i]
            if self.dataset_B.loc[self.dataset_B['index']==pair[0], 'celltype'].astype(str).values ==\
                self.dataset_A.loc[self.dataset_A['index']==pair[1],'celltype'].astype(str).values:
                if 'false' in self.mode:
                    continue
            if not self.conf is None:
                if 'low' in self.mode and not self.conf[i]: 
                    continue
            point0 = np.append(self.dataset_A[self.dataset_A['index']==pair[1]][['x','y']], 0)
            point1 = np.append(self.dataset_B[self.dataset_B['index']==pair[0]][['x','y']], 1)
            coord = np.row_stack((point0,point1))
            ax.scatter(point0[0],point0[1],point0[2],color='red',alpha=1,s=0.3)
            ax.scatter(point1[0],point1[1],point1[2],color='red',alpha=1,s=0.3)
            ax.plot(coord[:,0], coord[:,1], coord[:,2], color=self.highlight_color,
                    linestyle="dashed",linewidth=line_width,alpha=line_alpha)


class match_3D_celltype(match_3D_multi):
    r"""
    Highlight the celltype mapping, child of class:`match_3D_multi()`
    
    Parameters
    ---------
    dataset_A
        pandas dataframe which contain ['index','x','y']
    dataset_B
        pandas dataframe which contain ['index','x','y']
    matching
        matching results
    highlight_celltype
        celltypes to highlight in two datasets
    highlight_line
        color to highlight the line
    highlight_cell
        color to highlight the cell
    meta
        dataframe col name of meta, such as celltype
    expr
        dataframe col name of gene expr
    subsample_size
        subsample size of matches
    reliability
        if the match is reliable
    scale_coordinate
        if scale the coordinate via `data - np.min(data)) / (np.max(data) - np.min(data))` 
    rotate
        how to rotate the slides (force scale_coordinate)
    change_xy
        exchange x and y on dataset_B
    subset
        index of query cells to be plotted
        
    Note
    ----------
    dataset_A and dataset_B can in different length
    """
    def __init__(self,dataset_A: pd.DataFrame,
                dataset_B: pd.DataFrame,
                matching: np.ndarray,
                highlight_celltype: Optional[List[List[str]]]=[[],[]],
                highlight_line: Optional[Union[List[str],str]]='red',
                highlight_cell: Optional[str]=None,
                meta: Optional[str]=None,
                expr: Optional[str]=None,
                subsample_size: Optional[int]=300,
                reliability: Optional[np.ndarray]=None,
                scale_coordinate: Optional[bool]=False,
                rotate: Optional[List[str]]=None,
                exchange_xy: Optional[bool]=False,
                subset: Optional[Union[np.ndarray,List[int]]]=None
        ) -> None:
        super(match_3D_celltype, self).__init__(dataset_A,dataset_B,matching,meta,expr,subsample_size,
                                                reliability,scale_coordinate,rotate,exchange_xy,subset)
        assert set(highlight_celltype[0]).issubset(set(self.celltypes))
        assert set(highlight_celltype[1]).issubset(set(self.celltypes))
        self.highlight_celltype = highlight_celltype
        self.highlight_line = highlight_line
        self.highlight_cell = highlight_cell
        
    def draw_lines(self, ax, show_error, show_celltype, default_color, line_width:float=0.3, line_alpha:float=0.7)-> None:
        if len(self.highlight_celltype[0]) >= len(self.highlight_celltype[1]):
            color_index = self.highlight_celltype[0]
        else:
            color_index = self.highlight_celltype[1]
        if type(self.highlight_line) == list and len(self.highlight_line) >= len(color_index):
            cmap = self.highlight_line
        else:
            cmap = get_color(len(color_index))
        
        for i in range(self.matching.shape[1]):
            pair = self.matching[:,i]
            a = self.dataset_A.loc[self.dataset_A['index']==pair[1], self.meta].astype(str).values
            b = self.dataset_B.loc[self.dataset_B['index']==pair[0], self.meta].astype(str).values
            if a not in self.highlight_celltype[0] or b not in self.highlight_celltype[1]:
                continue
            point0 = np.append(self.dataset_A[self.dataset_A['index']==pair[1]][['x','y']], 0)
            point1 = np.append(self.dataset_B[self.dataset_B['index']==pair[0]][['x','y']], 1)
            coord = np.row_stack((point0,point1))
            if self.highlight_cell:
                ax.scatter(point0[0],point0[1],point0[2],color=self.highlight_cell,alpha=1,s=1)
                ax.scatter(point1[0],point1[1],point1[2],color=self.highlight_cell,alpha=1,s=1)
            if isinstance(cmap, list):
                color = cmap[color_index.index(a)] if len(self.highlight_celltype[0]) >= len(self.highlight_celltype[1]) else cmap[color_index.index(b)]
            else:
                color = cmap
            color = color if show_error else self.highlight_line
            ax.plot(coord[:,0], coord[:,1], coord[:,2], color=color, linestyle="dashed", linewidth=line_width, alpha=line_alpha)


def Sankey(matching_table:pd.DataFrame,
           filter_num:Optional[int]=50,
           color:Optional[List[str]]='red',
           title:Optional[str]='',
           prefix:Optional[List[str]]=['E11.5','E12.5'],
           layout:Optional[List[int]]=[1300,900],
           font_size:Optional[float]=15,
           font_color:Optional[str]='Black',
           save_name:Optional[str]=None,
           format:Optional[str]='png',
           width:Optional[int]=1200,
           height:Optional[int]=1000,
           return_fig:Optional[bool]=False
    ) -> None:
    r"""
    Sankey plot of celltype
    
    Parameters
    ----------
    matching_tables
        list of matching table
    filter_num
        filter number of matches
    color
        color of node
    title
        plot title
    prefix
        prefix to distinguish datasets
    layout
        layout size of picture
    font_size
        font size in plot
    font_color
        font color in plot
    save_name
        save file name (None for not save)
    format
        save picture format (see https://plotly.com/python/static-image-export/ for more details)
    width
        save picture width
    height
        save picture height
    return_fig
        if return plotly figure
    """
    source, target, value = [], [], []
    label_ref = [a + f'_{prefix[0]}' for a in matching_table.columns.to_list()]
    label_query = [a + f'_{prefix[1]}' for a in matching_table.index.to_list()]
    label_all = label_query + label_ref
    label2index = dict(zip(label_all, list(range(len(label_all)))))
    
    for i, query in enumerate(label_query):
        for j, ref in enumerate(label_ref):
            if int(matching_table.iloc[i,j]) > filter_num:
                target.append(label2index[query])
                source.append(label2index[ref])
                value.append(int(matching_table.iloc[i,j]))

    fig = go.Figure(
                    data=[go.Sankey(
                                    node = dict(pad = 50,
                                                thickness = 50,
                                                line = dict(color = "green", width = 0.5),
                                                label = label_all,
                                                color = color),
                                    link = dict(source = source, # indices correspond to labels, eg A1, A2, A1, B1, ...
                                                target = target,
                                                value = value)
                                    )
                        ],
                    layout=go.Layout(autosize=False, width=layout[0], height=layout[1])
                   )

    fig.update_layout(title_text=title, font_size=font_size, font_color=font_color)
    fig.show()
    if save_name != None:
        fig.write_image(save_name + f'.{format}', width=width, height=height)
    if return_fig:
        return fig
    

def multi_Sankey(matching_tables:List[pd.DataFrame],
                color:Optional[List[str]]='random',
                title:Optional[str]='Sankey plot',
                layout:Optional[List[int]]=[1300,900],
                day:Optional[float]=0,
                save_name:Optional[str]=None,
                format:Optional[str]='svg',
    ) -> None:
    r"""
    Sankey plot of celltype in multi datasets
    
    Parameters
    ----------
    matching_tables
        list of matching table
    color
        how to color the nodes, 'random' for random color, 'celltype' for color by celltype
    title
        plot title
    layout
        layout size of picture
    day
        start day of dataset for temporal order
    save_name
        save file name (None for not save)
    format
        saved picture format (see https://plotly.com/python/static-image-export/ for more details)
    """
    mappings = len(matching_tables) + 1
    prefixes = [day + i for i in range(mappings)]
    source, target, value, label_all= [], [], [], set()
    for i, matching_table in enumerate(matching_tables):
        label_ref = [a + f'_{prefixes[i]}' for a in matching_table.columns.to_list()]
        label_query = [a + f'_{prefixes[i+1]}' for a in matching_table.index.to_list()]
        # label_all.add(label_ref)
        for i in label_ref + label_query:
            label_all.add(i) 
    label2index = dict(zip(label_all, list(range(len(label_all)))))
    
    for matching_table,prefix in zip(matching_tables,prefixes):
        for i, query in enumerate(matching_table.index):
            for j, ref in enumerate(matching_table.columns):
                if int(matching_table.iloc[i,j]) > 10:
                    target.append(label2index[query+'_'+str(prefix+1)])
                    source.append(label2index[ref+'_'+str(prefix)])
                    value.append(int(matching_table.iloc[i,j]))
                    
    if color == 'random':
        color = [get_color()] * matching_tables[0].shape[0]
        for matching_table in matching_tables:
            color += [get_color()] * matching_table.shape[1]
    elif color == 'celltype':
        pass

    fig = go.Figure(
                    data=[go.Sankey(node = dict(
                                                pad = 50,
                                                thickness = 50,
                                                line = dict(color="green", width=0.5),
                                                label = list(label_all),
                                                color = color
                                                ),
                                    link = dict(source = source, # indices correspond to labels, eg A1, A2, A1, B1, ...
                                                target = target,
                                                value = value)
                                    )
                          ],
                    layout=go.Layout(autosize=False, width=layout[0], height=layout[1])
                   )

    fig.update_layout(title_text=title, font_size=10)
    if save_name != None:
        fig.write_image(save_name + f'.{format}', width=layout[0], height=layout[1])
    fig.show()
    

def matching_2d(matching:np.ndarray,
                ref:AnnData,
                src:AnnData,
                biology_meta:str,
                topology_meta:str,
                spot_size:Optional[int]=5,
                title:Optional[str]="2D matching",
                save:Optional[str]=None
    ) -> None:
    r"""
    Visualize the matching result in 2D space
    
    Parameters
    ----------
    matching
        matching result
    ref
        reference dataset
    src
        target dataset
    biology_meta
        celltype meta colname of adata.obs
    topology_meta
        region meta colname of adata.obs
    spot_size
        size of spot for visualization
    title
        plot title
    save
        save file name (None for not save)
    """
    src.obs['target_celltype'] = ref.obs.iloc[matching[1,:],:][biology_meta].to_list()
    src.obs['target_region'] = ref.obs.iloc[matching[1,:],:][topology_meta].to_list()
    src.obs["vis"] = 'celltype_false_region_false'
    src.obs["vis"] = src.obs["vis"].astype('str')
    
    cell_type_match = src.obs[biology_meta] == src.obs['target_celltype']
    region_match = src.obs[topology_meta] == src.obs['target_region']
    cell_type_match = cell_type_match.to_numpy()
    region_match = region_match.to_numpy()
    
    src.obs.loc[np.logical_and(cell_type_match, region_match), 'vis'] = 'celltype_true_region_true'
    src.obs.loc[np.logical_and(~cell_type_match, region_match), 'vis'] = 'celltype_false_region_true'
    src.obs.loc[np.logical_and(cell_type_match, ~region_match), 'vis'] = 'celltype_true_region_false'
    
    sc.pl.spatial(src, color="vis", spot_size=spot_size, title=title,
                  palette=['red', 'purple', 'yellow','green'],
                  save=save)
    
    del src.obs['target_celltype']
    del src.obs['target_region']
    del src.obs['vis']

import matplotlib.colors as mcolors
def html_to_rgba_string(html_color):
    rgba = mcolors.to_rgba(html_color)
    # 将 RGBA 转换为字符串格式 'rgba(r, g, b, a)'
    rgba_string = f'rgba({int(rgba[0]*255)}, {int(rgba[1]*255)}, {int(rgba[2]*255)}, {rgba[3]})'
    return rgba_string

def Sankey_multi(adata_li,prefix_li,matching_li,
                clusters='annotation',filter_num=10,
                layout=[1300,900],node_opacity = 0.8,
                 link_opacity = 0.2,
                 title='',
           font_size=15,
           font_color='Black',
           save_name=None,
           format='png',
           width=1200,
           height=1000,
           return_fig=False):
    #1. get adata_df
    adata_df={}
    for ad,prefix in zip(adata_li,prefix_li):
        adata_df[prefix]=pd.DataFrame({'index':range(ad.shape[0]),
                          'x': ad.obsm['spatial'][:,0],
                          'y': ad.obsm['spatial'][:,1],
                          'celltype':ad.obs[clusters]})
    
    #2. target_celltype
    matching_table_li=[]
    for idx,prefix in enumerate(prefix_li):
        if idx==0:
            prefix_pre=prefix
            continue
        adata_df[prefix]['target_celltype'] = adata_df[prefix_pre].iloc[matching_li[idx-1][1,:],:]['celltype'].to_list()
        matching_table_li.append(adata_df[prefix].groupby(['celltype', 'target_celltype']).size().unstack(fill_value=0))
        prefix_pre=prefix
        
    #3. Sankey prepare
    source, target, value = [], [], []
    label_all=[]
    color_all=[]
    for ad,prefix in zip(adata_li,
                       prefix_li):
        label_all=label_all+list(set([i+'_'+prefix for i in ad.obs[clusters].tolist()]))


    anno_key=clusters
    color_dict={}
    for ad,prefix in zip(adata_li,
                       prefix_li):
        if '{}_colors'.format(anno_key) in ad.uns.keys():
            c_map=dict(zip(ad.obs[anno_key].cat.categories.tolist(),
                                        ad.uns['{}_colors'.format(anno_key)]))
            for cname in c_map.keys():
                color_dict[cname+'_'+prefix]=c_map[cname]


    color_all=[html_to_rgba_string(color_dict[i]) for i in label_all]
    color_all_dict=dict(zip(label_all, color_all))
    label2index = dict(zip(label_all, list(range(len(label_all)))))
    
    color_all=[
        f'rgba(255,0,255,{node_opacity})' 
        if c == "magenta" else c.replace('1.0', str(node_opacity)) 
        for c in color_all]
    
    #filter_num=10
    #matching_table_li=[matching_table1,matching_table2]
    paired_list = [[prefix_li[i], prefix_li[i+1]] for i in range(len(prefix_li)-1)]

    for matching_t,pre in zip(matching_table_li,paired_list):
        for i, query in enumerate(matching_t.index.tolist()):
            for j, ref in enumerate(matching_t.columns.tolist()):
                if int(matching_t.loc[query,ref]) > filter_num:
                    #print(1,query+'_'+pre[1],ref+'_'+pre[0])
                    if (query+'_'+pre[1] in label2index.keys()) and (ref+'_'+pre[0] in label2index.keys() ):
                        #print(query+'_'+pre[1],ref+'_'+pre[0])
                        target.append(label2index[query+'_'+pre[1]])
                        source.append(label2index[ref+'_'+pre[0]])
                        value.append(int(matching_t.iloc[i,j]))

    
    link_color = [color_all[src].replace(str(node_opacity), str(link_opacity))
                                    for src in source]
    
    fig = go.Figure(data=[go.Sankey(
                valueformat = ".0f",
                valuesuffix = "TWh",
                # Define nodes
                node = dict(
                  pad = 15,
                  thickness = 15,
                  line = dict(color = "black", width = 0.5),
                  label =  label_all,
                  color =  color_all
                ),
                # Add links
                link = dict(
                  source =  source,
                  target =  target,
                  value =  value,
                  label =  label_all,
                  color =  link_color
            ))],
                   layout=go.Layout(autosize=False, width=layout[0], height=layout[1])
                   )

    fig.update_layout(title_text=title, font_size=font_size, font_color=font_color)
    fig.show()
    if save_name != None:
        fig.write_image(save_name + f'.{format}', width=width, height=height)
    if return_fig:
        return fig


from __future__ import annotations

import logging
import os

import numpy as np
import pandas as pd
import xgboost as xgb

from popv import settings
from popv.algorithms._base_algorithm import BaseAlgorithm


class XGboost(BaseAlgorithm):
    """
    Class to compute XGboost classifier.

    Parameters
    ----------
    batch_key
        Key in obs field of adata for batch information.
    labels_key
        Key in obs field of adata for cell-type information.
    layer_key
        Key in layers field of adata used for classification. By default uses 'X' (log1p10K).
    result_key
        Key in obs in which celltype annotation results are stored.
    classifier_dict
        Dictionary to supply non-default values for XGboost classifier.
        Options at :func:`xgboost.train`.
        Default is {'tree_method': 'hist', 'device': 'cuda' if settings.cuml else 'cpu', 'objective': 'multi:softprob'}.
    """

    def __init__(
        self,
        batch_key: str | None = "_batch_annotation",
        labels_key: str | None = "_labels_annotation",
        layer_key: str | None = None,
        result_key: str | None = "popv_xgboost_prediction",
        classifier_dict: dict | None = None,
    ) -> None:
        super().__init__(batch_key=batch_key, labels_key=labels_key, result_key=result_key)
        self.layer_key = layer_key
        self.classifier_dict = {
            "tree_method": "hist",
            "device": "cuda" if settings.cuml else "cpu",
            "objective": "multi:softprob",
        }
        if classifier_dict:
            self.classifier_dict.update(classifier_dict)

    def predict(self, adata):
        """
        Predict celltypes using XGboost.

        Parameters
        ----------
        adata
            Anndata object. Results are stored in adata.obs[self.result_key].
        """
        logging.info(f'Computing XGboost classifier. Storing prediction in adata.obs["{self.result_key}"]')
        
        # Create a boolean mask for relabeling cells
        relabel_mask = adata.obs["_predict_cells"] == "relabel"
        subset = adata[relabel_mask]
        
        # Select layer if provided, otherwise use adata.X
        test_x = subset.layers[self.layer_key] if self.layer_key else subset.X
        test_y = subset.obs[self.labels_key].cat.codes.to_numpy()

        dtest = xgb.DMatrix(test_x, label=test_y)

        model_path = os.path.join(adata.uns["_save_path_trained_models"], "xgboost_classifier.model")
        if adata.uns["_prediction_mode"] == "retrain":
            train_idx = adata.obs["_ref_subsample"]
            training_data = adata[train_idx]
            train_x = training_data.layers[self.layer_key] if self.layer_key else training_data.X
            train_y = adata.obs.loc[train_idx, self.labels_key].cat.codes.to_numpy()
            dtrain = xgb.DMatrix(train_x, label=train_y)

            self.classifier_dict["num_class"] = len(adata.uns["label_categories"])
            bst = xgb.train(self.classifier_dict, dtrain, num_boost_round=300)
            bst.save_model(model_path)
        else:
            bst = xgb.Booster({"device": "cuda" if False else "cpu"})
            bst.load_model(model_path)

        output_probabilities = bst.predict(dtest)
        # Zero the probability for the unknown celltype
        unknown_idx = list(adata.uns["label_categories"]).index(adata.uns["unknown_celltype_label"])
        output_probabilities[:, unknown_idx] = 0.0

        # Map predictions back to cell-type labels
        predicted_labels = adata.uns["label_categories"][np.argmax(output_probabilities, axis=1)]
        if self.result_key not in adata.obs.columns:
            adata.obs[self.result_key] = adata.uns["unknown_celltype_label"]
        adata.obs.loc[relabel_mask, self.result_key] = predicted_labels

        # Optionally assign maximum prediction probability per cell
        if self.return_probabilities:
            prob_key = f"{self.result_key}_probabilities"
            if prob_key not in adata.obs.columns:
                adata.obs[prob_key] = pd.Series(dtype="float64")
            adata.obs.loc[relabel_mask, prob_key] = np.max(output_probabilities, axis=1).astype(float)



r"""
Pyomic (A omic framework for multi-omic analysis)
"""

try:
    from importlib.metadata import version
except ModuleNotFoundError:
    from pkg_resources import get_distribution
    version = lambda name: get_distribution(name).version

# Core submodules
from . import bulk, single, utils, bulk2single, pp, space, pl

# Optional PopV module requires ``scvi-tools``. Skip if dependency missing.

#usually
from .utils._data import read
from .utils._plot import palette,ov_plot_set,plot_set


name = "omicverse"
try:
    __version__ = version(name)
except Exception:
    __version__ = "unknown"


from ._settings import settings,generate_reference_table


# 导入 matplotlib.pyplot
import matplotlib.pyplot as plt

# 将 plt 作为 omicverse 的一个属性
plt = plt  # 注意：确保没有其他变量名冲突

import numpy as np

np = np  # 注意：确保没有其他变量名冲突

# 导入 pandas
import pandas as pd

# 将 pd 作为 omicverse 的一个属性
pd = pd  # 注意：确保没有其他变量名冲突


#! python
# -*- coding: utf-8 -*-

import os, sys, logging, json, glob
from collections import OrderedDict
from tempfile import TemporaryDirectory

from numpy import log, exp
import numpy as np
import pandas as pd
import requests
from .algorithm import enrichment_score, gsea_compute, ranking_metric
from .algorithm import enrichment_score_tensor, gsea_compute_tensor
from .parser import gsea_edb_parser, gsea_cls_parser
from .plot import gseaplot, heatmap
from .utils import mkdirs, log_init, retry, DEFAULT_LIBRARY, DEFAULT_CACHE_PATH


class GSEAbase(object):
    """base class of GSEA."""
    def __init__(self):
        self.outdir='temp_gseapy'
        self.gene_sets='KEGG_2016'
        self.fdr=0.05
        self.module='base'
        self.results=None
        self.res2d=None
        self.ranking=None
        self.ascending=False
        self.verbose=False
        self._processes=1
        self._logger=None
        self.ENRICHR_URL = "http://maayanlab.cloud"

    def prepare_outdir(self):
        """create temp directory."""
        self._outdir = self.outdir
        if self._outdir is None:
            self._tmpdir = TemporaryDirectory()
            self.outdir = self._tmpdir.name
        elif isinstance(self.outdir, str):
            mkdirs(self.outdir)
        else:
            raise Exception("Error parsing outdir: %s"%type(self.outdir))

        # handle gmt type
        if isinstance(self.gene_sets, str):
            _gset = os.path.split(self.gene_sets)[-1].lower().rstrip(".gmt")
        elif isinstance(self.gene_sets, dict):
            _gset = "blank_name"
        else:
            raise Exception("Error parsing gene_sets parameter for gene sets")

        logfile = os.path.join(self.outdir, "gseapy.%s.%s.log" % (self.module, _gset))
        return logfile

    def _set_cores(self):
        """set cpu numbers to be used"""

        cpu_num = os.cpu_count()-1
        if self._processes > cpu_num:
            cores = cpu_num
        elif self._processes < 1:
            cores = 1
        else:
            cores = self._processes
        # have to be int if user input is float
        self._processes = int(cores)

    def _load_ranking(self, rnk):
        """Parse ranking file. This file contains ranking correlation vector( or expression values)
           and gene names or ids.

            :param rnk: the .rnk file of GSEA input or a Pandas DataFrame, Series instance.
            :return: a Pandas Series with gene name indexed rankings

        """
        # load data
        if isinstance(rnk, pd.DataFrame):
            rank_metric = rnk.copy()
            # handle dataframe with gene_name as index.
            if rnk.shape[1] == 1: rank_metric = rnk.reset_index()
        elif isinstance(rnk, pd.Series):
            rank_metric = rnk.reset_index()
        elif os.path.isfile(rnk):
            rank_metric = pd.read_csv(rnk, header=None, sep="\t")
        else:
            raise Exception('Error parsing gene ranking values!')
        # sort ranking values from high to low
        rank_metric.sort_values(by=rank_metric.columns[1], ascending=self.ascending, inplace=True)
        # drop na values
        if rank_metric.isnull().any(axis=1).sum() >0:
            self._logger.warning("Input gene rankings contains NA values(gene name and ranking value), drop them all!")
            # print out NAs
            NAs = rank_metric[rank_metric.isnull().any(axis=1)]
            self._logger.debug('NAs list:\n'+NAs.to_string())
            rank_metric.dropna(how='any', inplace=True)
        # drop duplicate IDs, keep the first
        if rank_metric.duplicated(subset=rank_metric.columns[0]).sum() >0:
            self._logger.warning("Input gene rankings contains duplicated IDs, Only use the duplicated ID with highest value!")
            # print out duplicated IDs.
            dups = rank_metric[rank_metric.duplicated(subset=rank_metric.columns[0])]
            self._logger.debug('Dups list:\n'+dups.to_string())
            rank_metric.drop_duplicates(subset=rank_metric.columns[0], inplace=True, keep='first')
        # reset ranking index, because you have sort values and drop duplicates.
        rank_metric.reset_index(drop=True, inplace=True)
        rank_metric.columns = ['gene_name','rank']
        rankser = rank_metric.set_index('gene_name')['rank']
        self.ranking = rankser
        # return series
        return rankser

    def load_gmt(self, gene_list, gmt):
        """load gene set dict"""

        if isinstance(gmt, dict):
            genesets_dict = gmt
        elif isinstance(gmt, str):
            genesets_dict = self.parse_gmt(gmt)
        else:
            raise Exception("Error parsing gmt parameter for gene sets")
        
        subsets = list(genesets_dict.keys())
        self.n_genesets = len(subsets)
        for subset in subsets:
            subset_list = genesets_dict.get(subset)
            if isinstance(subset_list, set):
                subset_list = list(subset_list)
                genesets_dict[subset] = subset_list
            tag_indicator = np.in1d(gene_list, subset_list, assume_unique=True)
            tag_len = tag_indicator.sum()
            if  (self.min_size <= tag_len <= self.max_size) and tag_len < len(gene_list): 
                # tag_len should < gene_list
                continue
            del genesets_dict[subset]

        filsets_num = len(subsets) - len(genesets_dict)
        self._logger.info("%04d gene_sets have been filtered out when max_size=%s and min_size=%s"%(filsets_num, self.max_size, self.min_size))

        if filsets_num == len(subsets):
            self._logger.error("No gene sets passed through filtering condition!!!, try new parameters again!\n" +\
                               "Note: check gene name, gmt file format, or filtering size." )
            raise Exception("No gene sets passed through filtering condition")

        self._gmtdct=genesets_dict
        return genesets_dict

    def parse_gmt(self, gmt):
        """gmt parser"""

        if gmt.lower().endswith(".gmt"):
            with open(gmt) as genesets:
                 genesets_dict = { line.strip().split("\t")[0]: line.strip().split("\t")[2:]
                                  for line in genesets.readlines()}
            return genesets_dict

        elif gmt in DEFAULT_LIBRARY:
            pass
        elif gmt in self.get_libraries():
            pass
        else:
            self._logger.error("No supported gene_sets: %s"%gmt)
            raise Exception("No supported gene_sets: %s"%gmt)

        tmpname = "enrichr." + gmt + ".gmt"
        tempath = os.path.join(DEFAULT_CACHE_PATH, tmpname)
        # if file already download
        if os.path.isfile(tempath):
            self._logger.info("Enrichr library gene sets already downloaded in: %s, use local file"%DEFAULT_CACHE_PATH)
            return self.parse_gmt(tempath)
        else:
            return self._download_libraries(gmt)

    def get_libraries(self):
        """return active enrichr library name.Offical API """

        lib_url=self.ENRICHR_URL+'/Enrichr/datasetStatistics' 
        response = requests.get(lib_url, verify=True)
        if not response.ok:
            raise Exception("Error getting the Enrichr libraries")
        libs_json = json.loads(response.text)
        libs = [lib['libraryName'] for lib in libs_json['statistics']]
 
        return sorted(libs)

    def _download_libraries(self, libname):
        """ download enrichr libraries."""
        self._logger.info("Downloading and generating Enrichr library gene sets......")
        s = retry(5)
        # queery string
        ENRICHR_URL = self.ENRICHR_URL+'/Enrichr/geneSetLibrary'
        query_string = '?mode=text&libraryName=%s'
        # get
        response = s.get( ENRICHR_URL + query_string % libname, timeout=None)
        if not response.ok:
            raise Exception('Error fetching enrichment results, check internet connection first.')
        # reformat to dict and save to disk
        mkdirs(DEFAULT_CACHE_PATH)
        genesets_dict = {}
        outname = "enrichr.%s.gmt"%libname
        gmtout = open(os.path.join(DEFAULT_CACHE_PATH, outname), "w")
        for line in response.iter_lines(chunk_size=1024, decode_unicode='utf-8'):
            line=line.strip()
            k = line.split("\t")[0]
            v = list(map(lambda x: x.split(",")[0], line.split("\t")[2:]))
            genesets_dict.update({ k: v})
            outline = "%s\t\t%s\n"%(k, "\t".join(v))
            gmtout.write(outline)
        gmtout.close()

        return genesets_dict

    def _heatmat(self, df, classes, pheno_pos, pheno_neg):
        """only use for gsea heatmap"""
        
        cls_booA =list(map(lambda x: True if x == pheno_pos else False, classes))
        cls_booB =list(map(lambda x: True if x == pheno_neg else False, classes))
        datA = df.loc[:, cls_booA]
        datB = df.loc[:, cls_booB]
        datAB=pd.concat([datA,datB], axis=1)
        self.heatmat = datAB
        return

    def _plotting(self, rank_metric, results, graph_num, outdir, 
                  format, figsize, pheno_pos='', pheno_neg=''):
        """ Plotting API.
            :param rank_metric: sorted pd.Series with rankings values.
            :param results: self.results
            :param data: preprocessed expression table

        """
        
        # no values need to be returned
        if self._outdir is None: return
        #Plotting
        top_term = self.res2d.index[:graph_num]
        # multi-threading
        #pool = Pool(self._processes)
        for gs in top_term:
            hit = results.get(gs)['hit_indices']
            NES = 'nes' if self.module != 'ssgsea' else 'es'
            term = gs.replace('/','_').replace(":","_")
            outfile = '{0}/{1}.{2}.{3}'.format(self.outdir, term, self.module, self.format)

            # if self.module != 'ssgsea' and results.get(gs)['fdr'] > 0.05:
            #     continue
            gseaplot(rank_metric=rank_metric, term=term, hit_indices=hit,
                      nes=results.get(gs)[NES], pval=results.get(gs)['pval'], 
                      fdr=results.get(gs)['fdr'], RES=results.get(gs)['RES'],
                      pheno_pos=pheno_pos, pheno_neg=pheno_neg, figsize=figsize,
                      ofname=outfile)

            if self.module == 'gsea':
                outfile2 = "{0}/{1}.heatmap.{2}".format(self.outdir, term, self.format)
                heatmat = self.heatmat.iloc[hit, :]
                width = np.clip(heatmat.shape[1], 4, 20)
                height = np.clip(heatmat.shape[0], 4, 20)
                heatmap(df=heatmat, title=term, ofname=outfile2, 
                        z_score=0, figsize=(width, height), 
                        xticklabels=True, yticklabels=True)


       
    def _save_results(self, zipdata, outdir, module, gmt, rank_metric, permutation_type):
        """reformat gsea results, and save to txt"""

        res = OrderedDict()
        for gs, gseale, ind, RES in zipdata:
            rdict = OrderedDict()
            rdict['es'] = gseale[0]
            rdict['nes'] = gseale[1]
            rdict['pval'] = gseale[2]
            rdict['fdr'] = gseale[3]
            rdict['geneset_size'] = len(gmt[gs])
            rdict['matched_size'] = len(ind)
            #reformat gene list.
            _genes = rank_metric.index.values[ind]
            rdict['genes'] = ";".join([ str(g).strip() for g in _genes ])
            
            if self.module != 'ssgsea':
                # extract leading edge genes
                if rdict['es'] > 0:
                    # RES -> ndarray, ind -> list
                    idx = RES.argmax()
                    ldg_pos = list(filter(lambda x: x<= idx, ind))
                elif rdict['es'] < 0:
                    idx = RES.argmin()
                    ldg_pos = list(filter(lambda x: x >= idx, ind))
                else:
                    ldg_pos = ind # es == 0 ?
                rdict['ledge_genes'] = ';'.join(list(map(str,rank_metric.iloc[ldg_pos].index)))
                
            rdict['RES'] = RES
            rdict['hit_indices'] = ind
            # save to one odict
            res[gs] = rdict
        # save
        self.results  = res
        # save to dataframe
        res_df = pd.DataFrame.from_dict(res, orient='index')
        res_df.index.name = 'Term'
        res_df.drop(['RES','hit_indices'], axis=1, inplace=True)
        res_df.sort_values(by=['fdr','pval'], inplace=True)
        self.res2d = res_df

        if self._outdir is None: return
        out = os.path.join(outdir,'gseapy.{b}.{c}.report.csv'.format(b=module, c=permutation_type))
        if self.module == 'ssgsea':
            out = out.replace(".csv",".txt")
            msg = "# normalize enrichment scores calculated by random permutation procedure (GSEA method)\n" +\
                  "# It's not proper for publication. Please check the original paper!\n"
            self._logger.warning(msg)
            with open(out, 'a') as f:
                f.write(msg)
                res_df.to_csv(f, sep='\t')
        else:
            res_df.to_csv(out)

        return

class GSEA(GSEAbase):
    """GSEA main tool"""
    def __init__(self, data, gene_sets, classes, outdir='GSEA_output',
                 min_size=15, max_size=500, permutation_num=1000,
                 weighted_score_type=1, permutation_type='gene_set',
                 method='log2_ratio_of_classes', ascending=False,
                 processes=1, figsize=(6.5,6), format='pdf', graph_num=20,
                 no_plot=False, seed=None, verbose=False):
        super(GSEA, self).__init__()
        self.data = data
        self.gene_sets=gene_sets
        self.classes=classes
        self.outdir=outdir
        self.permutation_type=permutation_type
        self.method=method
        self.min_size=min_size
        self.max_size=max_size
        self.permutation_num=int(permutation_num) if int(permutation_num) > 0 else 0
        self.weighted_score_type=weighted_score_type
        self.ascending=ascending
        self._processes=processes
        self.figsize=figsize
        self.format=format
        self.graph_num=int(graph_num)
        self.seed=seed
        self.verbose=bool(verbose)
        self.module='gsea'
        self.ranking=None
        self._noplot=no_plot
        # init logger
        logfile = self.prepare_outdir()
        self._logger = log_init(name='gseapy',
                                filename=logfile,
                                log_level=logging.INFO if self.verbose else logging.WARNING)

    def load_data(self, cls_vec):
        """pre-processed the data frame.new filtering methods will be implement here.
        """
        # read data in
        if isinstance(self.data, pd.DataFrame) :
            exprs = self.data.copy()
            # handle index is gene_names
            if exprs.index.dtype == 'O':
                exprs = exprs.reset_index()
        elif os.path.isfile(self.data) :
            # GCT input format?
            if self.data.endswith("gct"):
                exprs = pd.read_csv(self.data, skiprows=2, sep="\t")
            else:
                exprs = pd.read_csv(self.data, comment='#',sep="\t")
        else:
            raise Exception('Error parsing gene expression DataFrame!')

        #drop duplicated gene names
        if exprs.iloc[:,0].duplicated().sum() > 0:
            self._logger.warning("Warning: dropping duplicated gene names, only keep the first values")
            exprs.drop_duplicates(subset=exprs.columns[0], inplace=True) #drop duplicate gene_names.
        if exprs.isnull().any().sum() > 0:
            self._logger.warning("Warning: Input data contains NA, filled NA with 0")
            exprs.dropna(how='all', inplace=True) #drop rows with all NAs
            exprs = exprs.fillna(0)
        # set gene name as index
        exprs.set_index(keys=exprs.columns[0], inplace=True)
        # select numberic columns
        df = exprs.select_dtypes(include=[np.number])

        # in case the description column is numeric
        if len(cls_vec) == (df.shape[1] -1) :
            df = df.iloc[:,1:]
        # drop any genes which std ==0
        cls_dict = {k:v for k, v in zip(df.columns, cls_vec)}
        df_std =  df.groupby(by=cls_dict, axis=1).std()
        df =  df[~df_std.isin([0]).any(axis=1)]
        df = df + 0.00001 # we don't like zeros!!!
  
        return df, cls_dict

    def run(self):
        """GSEA main procedure"""
        assert self.method in ['signal_to_noise', 's2n', 'abs_signal_to_noise', 'abs_s2n',
                               't_test', 'ratio_of_classes', 'diff_of_classes', 'log2_ratio_of_classes']
        assert self.permutation_type in ["phenotype", "gene_set"]
        assert self.min_size <= self.max_size

        # Start Analysis
        self._logger.info("Parsing data files for GSEA.............................")
        # phenotype labels parsing
        phenoPos, phenoNeg, cls_vector = gsea_cls_parser(self.classes)
        # select correct expression genes and values.
        dat, cls_dict = self.load_data(cls_vector)
        # data frame must have length > 1
        assert len(dat) > 1
        # ranking metrics calculation.
        dat2 = ranking_metric(df=dat, method=self.method, pos=phenoPos, neg=phenoNeg,
                              classes= cls_dict, ascending=self.ascending)
        self.ranking = dat2
        # filtering out gene sets and build gene sets dictionary
        gmt = self.load_gmt(gene_list=dat2.index.values, gmt=self.gene_sets)

        self._logger.info("%04d gene_sets used for further statistical testing....."% len(gmt))
        self._logger.info("Start to run GSEA...Might take a while..................")
        # cpu numbers
        self._set_cores()
        # compute ES, NES, pval, FDR, RES
        dataset = dat if self.permutation_type =='phenotype' else dat2
        gsea_results,hit_ind,rank_ES, subsets = gsea_compute_tensor(data=dataset, gmt=gmt, n=self.permutation_num,
                                                             weighted_score_type=self.weighted_score_type,
                                                             permutation_type=self.permutation_type,
                                                             method=self.method,
                                                             pheno_pos=phenoPos, pheno_neg=phenoNeg,
                                                             classes=cls_vector, ascending=self.ascending,
                                                             processes=self._processes, seed=self.seed)
        
        self._logger.info("Start to generate GSEApy reports and figures............")
        res_zip = zip(subsets, list(gsea_results), hit_ind, rank_ES)
        self._save_results(zipdata=res_zip, outdir=self.outdir, module=self.module,
                                   gmt=gmt, rank_metric=dat2, permutation_type=self.permutation_type)

        # reorder datarame for heatmap
        self._heatmat(df=dat.loc[dat2.index], classes=cls_vector, 
                      pheno_pos=phenoPos, pheno_neg=phenoNeg)
        # Plotting
        if not self._noplot:
            self._plotting(rank_metric=dat2, results=self.results,
                           graph_num=self.graph_num, outdir=self.outdir,
                           figsize=self.figsize, format=self.format,
                           pheno_pos=phenoPos, pheno_neg=phenoNeg)

        self._logger.info("Congratulations. GSEApy ran successfully.................\n")
        if self._outdir is None:
            self._tmpdir.cleanup()

        return


class Prerank(GSEAbase):
    """GSEA prerank tool"""
    def __init__(self, rnk, gene_sets, outdir='GSEA_prerank',
                 pheno_pos='Pos', pheno_neg='Neg', min_size=15, max_size=500,
                 permutation_num=1000, weighted_score_type=1,
                 ascending=False, processes=1, figsize=(6.5,6), format='pdf',
                 graph_num=20, no_plot=False, seed=None, verbose=False):
        super(Prerank, self).__init__()
        self.rnk =rnk
        self.gene_sets=gene_sets
        self.outdir=outdir
        self.pheno_pos=pheno_pos
        self.pheno_neg=pheno_neg
        self.min_size=min_size
        self.max_size=max_size
        self.permutation_num=int(permutation_num) if int(permutation_num) > 0 else 0
        self.weighted_score_type=weighted_score_type
        self.ascending=ascending
        self.figsize=figsize
        self.format=format
        self.graph_num=int(graph_num)
        self.seed=seed
        self.verbose=bool(verbose)
        self.ranking=None
        self.module='prerank'
        self._processes=processes
        self._noplot=no_plot
        # init logger
        logfile = self.prepare_outdir()
        self._logger = log_init(name='gseapy',
                                filename=logfile,
                                log_level=logging.INFO if self.verbose else logging.WARNING)


    def run(self):
        """GSEA prerank workflow"""

        assert self.min_size <= self.max_size

        # parsing rankings
        dat2 = self._load_ranking(self.rnk)
        assert len(dat2) > 1

        # cpu numbers
        self._set_cores()
        # Start Analysis
        self._logger.info("Parsing data files for GSEA.............................")
        # filtering out gene sets and build gene sets dictionary
        gmt = self.load_gmt(gene_list=dat2.index.values, gmt=self.gene_sets)

        self._logger.info("%04d gene_sets used for further statistical testing....."% len(gmt))
        self._logger.info("Start to run GSEA...Might take a while..................")
        # compute ES, NES, pval, FDR, RES
        gsea_results, hit_ind,rank_ES, subsets = gsea_compute(data=dat2, n=self.permutation_num, gmt=gmt,
                                                              weighted_score_type=self.weighted_score_type,
                                                              permutation_type='gene_set', method=None,
                                                              pheno_pos=self.pheno_pos, pheno_neg=self.pheno_neg,
                                                              classes=None, ascending=self.ascending,
                                                              processes=self._processes, seed=self.seed)
        self._logger.info("Start to generate gseapy reports, and produce figures...")
        res_zip = zip(subsets, list(gsea_results), hit_ind, rank_ES)
        self._save_results(zipdata=res_zip, outdir=self.outdir, module=self.module,
                                   gmt=gmt, rank_metric=dat2, permutation_type="gene_sets")

        # Plotting
        if not self._noplot:
            self._plotting(rank_metric=dat2, results=self.results,
                           graph_num=self.graph_num, outdir=self.outdir,
                           figsize=self.figsize, format=self.format,
                           pheno_pos=self.pheno_pos, pheno_neg=self.pheno_neg)

        self._logger.info("Congratulations. GSEApy runs successfully................\n")
        if self._outdir is None:
            self._tmpdir.cleanup()

        return


class SingleSampleGSEA(GSEAbase):
    """GSEA extension: single sample GSEA"""
    def __init__(self, data, gene_sets, outdir="GSEA_SingleSample", sample_norm_method='rank',
                 min_size=15, max_size=2000, permutation_num=0, weighted_score_type=0.25,
                 scale=True, ascending=False, processes=1, figsize=(7,6), format='pdf',
                 graph_num=20, no_plot=True, seed=None, verbose=False):
        super(SingleSampleGSEA, self).__init__()
        self.data=data
        self.gene_sets=gene_sets
        self.outdir=outdir
        self.sample_norm_method=sample_norm_method
        self.weighted_score_type=weighted_score_type
        self.scale = scale
        self.min_size=min_size
        self.max_size=max_size
        self.permutation_num=int(permutation_num) if int(permutation_num) > 0 else 0
        self.ascending=ascending
        self.figsize=figsize
        self.format=format
        self.graph_num=int(graph_num)
        self.seed=seed
        self.verbose=bool(verbose)
        self.ranking=None
        self.module='ssgsea'
        self._processes=processes
        self._noplot=no_plot
        # init logger

        logfile = self.prepare_outdir()
        self._logger = log_init(name='gseapy',
                                filename=logfile,
                                log_level=logging.INFO if self.verbose else logging.WARNING)

    def corplot(self):
        """NES Correlation plot
        TODO
        """
    def setplot(self):
        """ranked genes' location plot
        TODO
        """

    def load_data(self):
        #load data
        exprs = self.data
        if isinstance(exprs, pd.DataFrame):
            rank_metric = exprs.copy()
            # handle dataframe with gene_name as index.
            self._logger.debug("Input data is a DataFrame with gene names")
            # handle index is not gene_names
            if rank_metric.index.dtype != 'O':
                rank_metric.set_index(keys=rank_metric.columns[0], inplace=True)

            rank_metric = rank_metric.select_dtypes(include=[np.number])
        elif isinstance(exprs, pd.Series):
            # change to DataFrame
            self._logger.debug("Input data is a Series with gene names")
            rank_metric = pd.DataFrame(exprs)
            # rename col if name attr is none
            rank_metric.columns = ["sample1"]
        elif os.path.isfile(exprs):
            # GCT input format?
            if exprs.endswith("gct"):
                rank_metric = pd.read_csv(exprs, skiprows=1, comment='#', index_col=0, sep="\t")
            else:
                # just txt file like input
                rank_metric = pd.read_csv(exprs, comment='#', index_col=0, sep="\t")
                if rank_metric.shape[1] ==1:
                    # rnk file like input
                    rank_metric = pd.read_csv(exprs, header=None, comment='#',
                                                names=['sample1'], index_col=0, sep="\t")
            # select numbers
            rank_metric = rank_metric.select_dtypes(include=[np.number])
        else:
            raise Exception('Error parsing gene ranking values!')

        if rank_metric.index.duplicated().sum() > 0:
            self._logger.warning("Warning: dropping duplicated gene names, only keep the first values")
            rank_metric = rank_metric.loc[rank_metric.index.drop_duplicates(keep='first')]
            rank_metric = rank_metric.loc[rank_metric.index.dropna()]
        if rank_metric.isnull().any().sum() > 0:
            self._logger.warning("Warning: Input data contains NA, filled NA with 0")
            rank_metric = rank_metric.fillna(0)

        return rank_metric

    def norm_samples(self, dat):
        """normalization samples
           see here: http://rowley.mit.edu/caw_web/ssGSEAProjection/ssGSEAProjection.Library.R
        """

        if self.sample_norm_method == 'rank':
            data = dat.rank(axis=0, method='average', na_option='bottom')
            data = 10000*data / data.shape[0]
        elif self.sample_norm_method == 'log_rank':
            data = dat.rank(axis=0, method='average', na_option='bottom')
            data = log(10000*data / data.shape[0] + exp(1))
        elif self.sample_norm_method == 'log':
            dat[dat < 1] = 1
            data = log(dat + exp(1))
        elif self.sample_norm_method == 'custom':
            self._logger.info("Use custom rank metric for ssGSEA")
            data = dat
        else:
            raise Exception("No supported method: %s"%self.sample_norm_method)

        return data

    def run(self):
        """run entry"""
        self._logger.info("Parsing data files for ssGSEA...........................")
        # load data
        data = self.load_data()
        # normalized samples, and rank
        normdat = self.norm_samples(data)
        # filtering out gene sets and build gene sets dictionary
        gmt = self.load_gmt(gene_list=normdat.index.values, gmt=self.gene_sets)
        self._logger.info("%04d gene_sets used for further statistical testing....."% len(gmt))
        # set cpu numbers
        self._set_cores()
        # start analysis
        self._logger.info("Start to run ssGSEA...Might take a while................")
        if self.permutation_num == 0 :
            # ssGSEA without permutation
            self.runSamples(df=normdat, gmt=gmt)
        else:
            # run permutation procedure and calculate pvals, fdrs
            self._logger.warning("run ssGSEA with permutation procedure, don't use these part of results for publication.")
            self.runSamplesPermu(df=normdat, gmt=gmt)
        # clean up all outputs if _outdir is None
        if self._outdir is None:
            self._tmpdir.cleanup()

    def runSamplesPermu(self, df, gmt=None):
        """Single Sample GSEA workflow with permutation procedure"""

        assert self.min_size <= self.max_size
        mkdirs(self.outdir)
        self.resultsOnSamples = OrderedDict()
        outdir = self.outdir
        # iter through each sample
        for name, ser in df.iteritems():
            self.outdir = os.path.join(outdir, str(name))
            self._logger.info("Run Sample: %s " % name)
            mkdirs(self.outdir)
            # sort ranking values from high to low or reverse
            dat2 = ser.sort_values(ascending=self.ascending)
            # reset integer index, or caused unwanted problems
            # df.reset_index(drop=True, inplace=True)

            # compute ES, NES, pval, FDR, RES
            gsea_results, hit_ind,rank_ES, subsets = gsea_compute(data=dat2, n=self.permutation_num, gmt=gmt,
                                                                  weighted_score_type=self.weighted_score_type,
                                                                  permutation_type='gene_set', method=None,
                                                                  pheno_pos='', pheno_neg='',
                                                                  classes=None, ascending=self.ascending,
                                                                  processes=self._processes,
                                                                  seed=self.seed, single=True, scale=self.scale)

            # write file
            res_zip = zip(subsets, list(gsea_results), hit_ind, rank_ES)
            self._save_results(zipdata=res_zip, outdir=self.outdir, module=self.module,
                                gmt=gmt, rank_metric=dat2, permutation_type="gene_sets")
            self.resultsOnSamples[name] = self.res2d.es
            # plotting
            if self._noplot: continue
            self._logger.info("Plotting Sample: %s \n" % name)
            self._plotting(rank_metric=dat2, results=self.results,
                           graph_num=self.graph_num, outdir=self.outdir,
                           figsize=self.figsize, format=self.format)

        # save es, nes to file
        self._save(outdir)

        return

    def runSamples(self, df, gmt=None):
        """Single Sample GSEA workflow.
           multiprocessing utility on samples.
        """

        # df.index.values are gene_names
        # Save each sample results to odict
        self.resultsOnSamples = OrderedDict()
        outdir = self.outdir
        # run ssgsea for gct expression matrix
        #multi-threading
        subsets = sorted(gmt.keys())
        tempdat=[]
        tempes=[]
        names=[]
        rankings=[]
        #pool = Pool(processes=self._processes)
        np.random.seed(self.seed)
        random_state = np.random.randint(np.iinfo(np.int32).max, size=df.shape[1])
        for name, ser in df.iteritems():
            #prepare input
            dat = ser.sort_values(ascending=self.ascending)
            tempdat.append(dat)
            rankings.append(dat)
            names.append(name)
        from joblib import delayed, Parallel
        tempes = Parallel(n_jobs=self._processes)(
                             delayed(enrichment_score_tensor)(
                                               dat.index.values, dat.values, gmt,
                                               self.weighted_score_type,
                                               self.permutation_num, rs, True,
                                               self.scale) 
                             for dat, rs in zip(tempdat, random_state))

        # save results and plotting
        for i, temp in enumerate(tempes):
            name, rnk = names[i], rankings[i]
            self._logger.info("Calculate Enrichment Score for Sample: %s "%name)
            # es, esnull, hit_ind, RES = temp.get()
            es, esnull, hit_ind, RES = temp
            # create results subdir
            self.outdir= os.path.join(outdir, str(name))
            mkdirs(self.outdir)
            # save results
            self.resultsOnSamples[name] = pd.Series(data=es, index=subsets, name=name)
            # plotting
            if self._noplot: continue
            self._logger.info("Plotting Sample: %s \n" % name)
            for i, term in enumerate(subsets):
                term = term.replace('/','_').replace(":","_")
                outfile = '{0}/{1}.{2}.{3}'.format(self.outdir, term, self.module, self.format)
                gseaplot(rank_metric=rnk, term=term, 
                         hit_indices=hit_ind[i], nes=es[i], pval=1, fdr=1, 
                         RES=RES[i], pheno_pos='', pheno_neg='', 
                         figsize=self.figsize, ofname=outfile)
        # save es, nes to file
        self._save(outdir)

        return

    def _save(self, outdir):
        """save es and stats"""
        # save raw ES to one csv file
        samplesRawES = pd.DataFrame(self.resultsOnSamples)
        samplesRawES.index.name = 'Term|ES'
        # normalize enrichment scores by using the entire data set, as indicated
        # by Barbie et al., 2009, online methods, pg. 2
        samplesNES = samplesRawES / (samplesRawES.values.max() - samplesRawES.values.min())
        samplesNES = samplesNES.copy()
        samplesNES.index.rename('Term|NES', inplace=True)
        self.res2d = samplesNES
        self._logger.info("Congratulations. GSEApy runs successfully................\n")
        if self._outdir is None: return
        # write es
        outESfile = os.path.join(outdir, "gseapy.samples.raw.es.txt")
        with open(outESfile, 'a') as f:
            if self.scale:
                f.write('# scale the enrichment scores by number of genes in the gene sets\n')
                f.write('# this normalization has not effects on the final NES, ' + \
                        'as indicated by Barbie et al., 2009, online methods, pg. 2\n')
            else:
                f.write('# raw enrichment scores of all data\n')
                f.write('# no scale es by numbers of genes in the gene sets\n')
            samplesRawES.to_csv(f, sep='\t')

        outNESfile = os.path.join(outdir, "gseapy.samples.normalized.es.txt")
        with open(outNESfile, 'a') as f:
            f.write('# normalize enrichment scores by using the entire data set\n')
            f.write('# as indicated by Barbie et al., 2009, online methods, pg. 2\n')
            samplesNES.to_csv(f, sep='\t')
        return

class Replot(GSEAbase):
    """To reproduce GSEA desktop output results."""
    def __init__(self, indir, outdir='GSEApy_Replot', weighted_score_type=1,
                  min_size=3, max_size=1000, figsize=(6.5,6), format='pdf', verbose=False):
        self.indir=indir
        self.outdir=outdir
        self.weighted_score_type=weighted_score_type
        self.min_size=min_size
        self.max_size=max_size
        self.figsize=figsize
        self.format=format
        self.verbose=bool(verbose)
        self.module='replot'
        self.gene_sets=None
        self.ascending=False
        # init logger
        mkdirs(self.outdir)
        outlog = os.path.join(self.outdir,"gseapy.%s.%s.log"%(self.module,"run"))
        self._logger = log_init(name='gseapy',
                                filename=logfile,
                                log_level=logging.INFO if self.verbose else logging.WARNING)
    def run(self):
        """main replot function"""
        assert self.min_size <= self.max_size

        # parsing files.......
        try:
            results_path = glob.glob(self.indir+'*/edb/results.edb')[0]
            rank_path =  glob.glob(self.indir+'*/edb/*.rnk')[0]
            gene_set_path =  glob.glob(self.indir+'*/edb/gene_sets.gmt')[0]
        except IndexError as e:
            raise Exception("Could not locate GSEA files in the given directory!")
        # extract sample names from .cls file
        cls_path = glob.glob(self.indir+'*/edb/*.cls')
        if cls_path:
            pos, neg, classes = gsea_cls_parser(cls_path[0])
        else:
            # logic for prerank results
            pos, neg = '',''
        # start reploting
        self.gene_sets = gene_set_path
        # obtain gene sets
        gene_set_dict = self.parse_gmt(gmt=gene_set_path)
        # obtain rank_metrics
        rank_metric = self._load_ranking(rank_path)
        correl_vector = rank_metric.values
        gene_list = rank_metric.index.values
        # extract each enriment term in the results.edb files and plot.
        
        database = gsea_edb_parser(results_path)
        for enrich_term, data in database.items():
            # extract statistical resutls from results.edb file
            hit_ind, nes, pval, fdr = data
            gene_set = gene_set_dict.get(enrich_term)
            # calculate enrichment score
            RES = enrichment_score(gene_list=gene_list, 
                                   correl_vector=correl_vector,
                                   gene_set=gene_set, 
                                   weighted_score_type=self.weighted_score_type,
                                   nperm=0)[-1]
            # plotting
            term = enrich_term.replace('/','_').replace(":","_")
            outfile = '{0}/{1}.{2}.{3}'.format(self.outdir, term, self.module, self.format)
            gseaplot(rank_metric=rank_metric, term=enrich_term, 
                         hit_indices=hit_ind, nes=nes, pval=pval, fdr=fdr, 
                         RES=RES, pheno_pos=pos, pheno_neg=neg, 
                         figsize=self.figsize, ofname=outfile)

        self._logger.info("Congratulations! Your plots have been reproduced successfully!\n")



def gsea(data, gene_sets, cls, outdir='GSEA_', min_size=15, max_size=500, permutation_num=1000,
          weighted_score_type=1,permutation_type='gene_set', method='log2_ratio_of_classes',
	      ascending=False, processes=1, figsize=(6.5,6), format='pdf',
          graph_num=20, no_plot=False, seed=None, verbose=False):
    """ Run Gene Set Enrichment Analysis.

    :param data: Gene expression data table, Pandas DataFrame, gct file.
    :param gene_sets: Enrichr Library name or .gmt gene sets file or dict of gene sets. Same input with GSEA.
    :param cls: A list or a .cls file format required for GSEA.
    :param str outdir: Results output directory.
    :param int permutation_num: Number of permutations for significance computation. Default: 1000.
    :param str permutation_type: Permutation type, "phenotype" for phenotypes, "gene_set" for genes.
    :param int min_size: Minimum allowed number of genes from gene set also the data set. Default: 15.
    :param int max_size: Maximum allowed number of genes from gene set also the data set. Default: 500.
    :param float weighted_score_type: Refer to :func:`algorithm.enrichment_score`. Default:1.
    :param method:  The method used to calculate a correlation or ranking. Default: 'log2_ratio_of_classes'.
                   Others methods are:

                   1. 'signal_to_noise'

                      You must have at least three samples for each phenotype to use this metric.
                      The larger the signal-to-noise ratio, the larger the differences of the means (scaled by the standard deviations);
                      that is, the more distinct the gene expression is in each phenotype and the more the gene acts as a “class marker.”

                   2. 't_test'

                      Uses the difference of means scaled by the standard deviation and number of samples.
                      Note: You must have at least three samples for each phenotype to use this metric.
                      The larger the tTest ratio, the more distinct the gene expression is in each phenotype
                      and the more the gene acts as a “class marker.”

                   3. 'ratio_of_classes' (also referred to as fold change).

                      Uses the ratio of class means to calculate fold change for natural scale data.

                   4. 'diff_of_classes'


                      Uses the difference of class means to calculate fold change for nature scale data


                   5. 'log2_ratio_of_classes'

                      Uses the log2 ratio of class means to calculate fold change for natural scale data.
                      This is the recommended statistic for calculating fold change for log scale data.


    :param bool ascending: Sorting order of rankings. Default: False.
    :param int processes: Number of Processes you are going to use. Default: 1.
    :param list figsize: Matplotlib figsize, accept a tuple or list, e.g. [width,height]. Default: [6.5,6].
    :param str format: Matplotlib figure format. Default: 'pdf'.
    :param int graph_num: Plot graphs for top sets of each phenotype.
    :param bool no_plot: If equals to True, no figure will be drawn. Default: False.
    :param seed: Random seed. expect an integer. Default:None.
    :param bool verbose: Bool, increase output verbosity, print out progress of your job, Default: False.

    :return: Return a GSEA obj. All results store to a dictionary, obj.results,
             where contains::

                 | {es: enrichment score,
                 |  nes: normalized enrichment score,
                 |  p: P-value,
                 |  fdr: FDR,
                 |  size: gene set size,
                 |  matched_size: genes matched to the data,
                 |  genes: gene names from the data set
                 |  ledge_genes: leading edge genes}


    """
    gs = GSEA(data, gene_sets, cls, outdir, min_size, max_size, permutation_num,
              weighted_score_type, permutation_type, method, ascending, processes,
               figsize, format, graph_num, no_plot, seed, verbose)
    gs.run()

    return gs


def ssgsea(data, gene_sets, outdir="ssGSEA_", sample_norm_method='rank', min_size=15, max_size=2000,
           permutation_num=0, weighted_score_type=0.25, scale=True, ascending=False, processes=1,
           figsize=(7,6), format='pdf', graph_num=20, no_plot=True, seed=None, verbose=False):
    """Run Gene Set Enrichment Analysis with single sample GSEA tool

    :param data: Expression table, pd.Series, pd.DataFrame, GCT file, or .rnk file format.
    :param gene_sets: Enrichr Library name or .gmt gene sets file or dict of gene sets. Same input with GSEA.
    :param outdir: Results output directory.
    :param str sample_norm_method: "Sample normalization method. Choose from {'rank', 'log', 'log_rank'}. Default: rank.

               1. 'rank': Rank your expression data, and transform by 10000*rank_dat/gene_numbers
               2. 'log' : Do not rank, but transform data by log(data + exp(1)), while data = data[data<1] =1.
               3. 'log_rank': Rank your expression data, and transform by log(10000*rank_dat/gene_numbers+ exp(1))
               4. 'custom': Do nothing, and use your own rank value to calculate enrichment score.
    
    see here: https://github.com/GSEA-MSigDB/ssGSEAProjection-gpmodule/blob/master/src/ssGSEAProjection.Library.R, line 86

    :param int min_size: Minimum allowed number of genes from gene set also the data set. Default: 15.
    :param int max_size: Maximum allowed number of genes from gene set also the data set. Default: 2000.
    :param int permutation_num: Number of permutations for significance computation. Default: 0.
    :param str weighted_score_type: Refer to :func:`algorithm.enrichment_score`. Default:0.25.
    :param bool scale: If True, normalize the scores by number of genes in the gene sets.
    :param bool ascending: Sorting order of rankings. Default: False.
    :param int processes: Number of Processes you are going to use. Default: 1.
    :param list figsize: Matplotlib figsize, accept a tuple or list, e.g. [width,height]. Default: [7,6].
    :param str format: Matplotlib figure format. Default: 'pdf'.
    :param int graph_num: Plot graphs for top sets of each phenotype.
    :param bool no_plot: If equals to True, no figure will be drawn. Default: False.
    :param seed: Random seed. expect an integer. Default:None.
    :param bool verbose: Bool, increase output verbosity, print out progress of your job, Default: False.

    :return: Return a ssGSEA obj. 
             All results store to  a dictionary, access enrichment score by obj.resultsOnSamples,
             and normalized enrichment score by obj.res2d.
             if permutation_num > 0, additional results contain::

                 | {es: enrichment score,
                 |  nes: normalized enrichment score,
                 |  p: P-value,
                 |  fdr: FDR,
                 |  size: gene set size,
                 |  matched_size: genes matched to the data,
                 |  genes: gene names from the data set
                 |  ledge_genes: leading edge genes, if permutation_num >0}


    """

    ss = SingleSampleGSEA(data, gene_sets, outdir, sample_norm_method, min_size, max_size,
                          permutation_num, weighted_score_type, scale, ascending,
                          processes, figsize, format, graph_num, no_plot, seed, verbose)
    ss.run()
    return ss


def prerank(rnk, gene_sets, outdir='GSEA_Prerank', pheno_pos='Pos', pheno_neg='Neg',
            min_size=15, max_size=500, permutation_num=1000, weighted_score_type=1,
            ascending=False, processes=1, figsize=(6.5,6), format='pdf',
            graph_num=20, no_plot=False, seed=None, verbose=False):
    """ Run Gene Set Enrichment Analysis with pre-ranked correlation defined by user.

    :param rnk: pre-ranked correlation table or pandas DataFrame. Same input with ``GSEA`` .rnk file.
    :param gene_sets: Enrichr Library name or .gmt gene sets file or dict of gene sets. Same input with GSEA.
    :param outdir: results output directory.
    :param int permutation_num: Number of permutations for significance computation. Default: 1000.
    :param int min_size: Minimum allowed number of genes from gene set also the data set. Default: 15.
    :param int max_size: Maximum allowed number of genes from gene set also the data set. Defaults: 500.
    :param str weighted_score_type: Refer to :func:`algorithm.enrichment_score`. Default:1.
    :param bool ascending: Sorting order of rankings. Default: False.
    :param int processes: Number of Processes you are going to use. Default: 1.
    :param list figsize: Matplotlib figsize, accept a tuple or list, e.g. [width,height]. Default: [6.5,6].
    :param str format: Matplotlib figure format. Default: 'pdf'.
    :param int graph_num: Plot graphs for top sets of each phenotype.
    :param bool no_plot: If equals to True, no figure will be drawn. Default: False.
    :param seed: Random seed. expect an integer. Default:None.
    :param bool verbose: Bool, increase output verbosity, print out progress of your job, Default: False.

    :return: Return a Prerank obj. All results store to  a dictionary, obj.results,
             where contains::

                 | {es: enrichment score,
                 |  nes: normalized enrichment score,
                 |  p: P-value,
                 |  fdr: FDR,
                 |  size: gene set size,
                 |  matched_size: genes matched to the data,
                 |  genes: gene names from the data set
                 |  ledge_genes: leading edge genes}


    """
    pre = Prerank(rnk, gene_sets, outdir, pheno_pos, pheno_neg,
                  min_size, max_size, permutation_num, weighted_score_type,
                  ascending, processes, figsize, format, graph_num, no_plot, seed, verbose)
    pre.run()
    return pre


def replot(indir, outdir='GSEA_Replot', weighted_score_type=1,
           min_size=3, max_size=1000, figsize=(6.5,6), format='pdf', verbose=False):
    """The main function to reproduce GSEA desktop outputs.

    :param indir: GSEA desktop results directory. In the sub folder, you must contain edb file folder.
    :param outdir: Output directory.
    :param float weighted_score_type: weighted score type. choose from {0,1,1.5,2}. Default: 1.
    :param list figsize: Matplotlib output figure figsize. Default: [6.5,6].
    :param str format: Matplotlib output figure format. Default: 'pdf'.
    :param int min_size: Min size of input genes presented in Gene Sets. Default: 3.
    :param int max_size: Max size of input genes presented in Gene Sets. Default: 5000.
                     You are not encouraged to use min_size, or max_size argument in :func:`replot` function.
                     Because gmt file has already been filtered.
    :param verbose: Bool, increase output verbosity, print out progress of your job, Default: False.

    :return: Generate new figures with selected figure format. Default: 'pdf'.

    """
    rep = Replot(indir, outdir, weighted_score_type,
                 min_size, max_size, figsize, format, verbose)
    rep.run()

    return



import scanpy as sc
import numpy as np

import torch
#from tensorflow_probability import math as tm
#tfk = tm.psd_kernels
#import spatial_factorization as sf
from typing import Optional
from scipy.sparse import csr_matrix
from ._townes_nsf_utils import *
from sklearn.decomposition import NMF

def construct_gems_using_pyliger(adata: sc.AnnData,
                                n_gems: int,
                                layer_key: str,
                                condition_key: str):
    import pyliger
    conditions = adata.obs[condition_key].unique().tolist()
    ad = adata.copy()

    ad.X = csr_matrix(ad.layers[layer_key].copy())
    ad.obs.index.name = 'index'
    ad.var.index.name = 'index'

    # Create LIGER object
    adata_list = []
    for cond in conditions:
        adata_cond =  adata[adata.obs[condition_key] == cond].copy()
        adata_cond.uns['sample_name'] = cond
        adata_list.append(adata_cond)

    adata_liger = pyliger.create_liger(adata_list, make_sparse=True)

    pyliger.normalize(adata_liger)
    pyliger.select_genes(adata_liger)
    pyliger.scale_not_center(adata_liger)

    # Save the var_names that were used for the NMF
    # adata_burkhardt.uns['pyliger_vars'] = burkhardt_liger.adata_list[0].var_names.tolist()

    pyliger.optimize_ALS(adata_liger, k = n_gems)

    # # Save results to adata
    X_gem = np.zeros((adata.n_obs, n_gems))
    pyliger_info = {}

    for i, cond in enumerate(conditions):
        cond_indices = np.where(adata.obs[condition_key] == cond)[0]
        X_gem[cond_indices] = adata_liger.adata_list[i].obsm['H']

        pyliger_info[cond] = {'H': adata_liger.adata_list[i].obsm['H'],
                            'W': adata_liger.adata_list[i].varm['W'],
                            'V': adata_liger.adata_list[i].varm['V']}

    adata.uns['pyliger_info'] = pyliger_info
    adata.uns['pyliger_info']['vars'] = adata_liger.adata_list[0].var_names.tolist()        
    adata.uns['pyliger_info']['n_gems'] =  n_gems
    adata.obsm['X_gem'] = X_gem

def construct_gems_using_nsf(adata: sc.AnnData,
                            n_gems: int,
                            layer_key: str,
                            spatial_key: str = "spatial",
                            n_inducing_pts: int = 500,
                            length_scale: float = 10.0):
    import spatial_factorization as sf
    import gpytorch
    ad = adata.copy()
    
    X = ad.obsm[spatial_key]
    # Take raw count data for NSF
    training_fraction = 1.0
    D, Dval = anndata_to_train_val(ad,
                                   layer=layer_key,
                                   train_frac=training_fraction,
                                   flip_yaxis=True)
    Ntr, J = D["Y"].shape
    Xtr = D["X"]
    ad = adata[:Ntr, :]
    
    # Convert to PyTorch tensors
    Dtf = prepare_datasets_torch(D, Dval=Dval)

    Z = kmeans_inducing_pts(Xtr, n_inducing_pts)
    M = Z.shape[0]  # number of inducing points

    # Define the Matern 3/2 kernel using gpytorch
    class MaternThreeHalvesKernel(gpytorch.kernels.MaternKernel):
        def __init__(self, **kwargs):
            super(MaternThreeHalvesKernel, self).__init__(nu=1.5, **kwargs)

    ker = MaternThreeHalvesKernel(lengthscale=length_scale)

    fit = sf.SpatialFactorization(J, n_gems, Z, psd_kernel=ker, length_scale=length_scale, nonneg=True, lik="poi")
    fit.init_loadings(D["Y"], X=Xtr, sz=D["sz"], shrinkage=0.3)
    tro = sf.ModelTrainer(fit)
    tro.train_model(*Dtf, status_freq=50)  # about 3 mins

    insf = interpret_nsf(fit, Xtr, S=100, lda_mode=False)

    adata.uns['nsf_info'] = insf
    adata.uns['nsf_info']['vars'] = adata.var_names.tolist()
    adata.uns['nsf_info']['n_gems'] = n_gems
    adata.obsm['X_gem'] = insf['factors']

# Helper function to convert datasets to PyTorch tensors
def prepare_datasets_torch(D, Dval=None):
    Dtorch = {
        "X": torch.tensor(D["X"], dtype=torch.float32),
        "Y": torch.tensor(D["Y"], dtype=torch.float32),
        "sz": torch.tensor(D["sz"], dtype=torch.float32)
    }
    if Dval is not None:
        Dtorch["Xval"] = torch.tensor(Dval["X"], dtype=torch.float32)
        Dtorch["Yval"] = torch.tensor(Dval["Y"], dtype=torch.float32)
        Dtorch["szval"] = torch.tensor(Dval["sz"], dtype=torch.float32)
    return Dtorch

def construct_gems_using_nmf(adata: sc.AnnData,
                                n_gems: int,
                                layer_key: str, 
                                random_state: int = 0,
                                max_iter: int = 1000):

    X_expr = adata.layers[layer_key].copy()
    
    model = NMF(n_components=n_gems, init='random', random_state=random_state, max_iter=max_iter)

    W = model.fit_transform(X_expr)
    H = model.components_

    W_sum = W.sum(axis=0)
    W_lda = W / W_sum

    H_scaled = H.T * W_sum
    H_sum = H_scaled.sum(axis=1)
    H_lda = (H_scaled.T / H_sum).T

    fact_orders = np.argsort(-H_lda.sum(axis=0))

    W_lda = W_lda[:, fact_orders]
    H_lda = H_lda[:, fact_orders].T

    adata.uns['nmf_info'] = {'n_gems': n_gems,
                             'vars': adata.var_names.tolist(),
                             'factors':W_lda,
                             'loadings':H_lda,
                             'totals':W_sum}

    adata.obsm['X_gem'] = W_lda



import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
import scanpy as sc

def embedding_atlas(adata,basis,color,
                    title=None,figsize=(4,4),ax=None,cmap='RdBu',
                    legend_loc = 'right margin',frameon='small',
                    fontsize=12):
    import scanpy as sc
    import pandas as pd
    import datashader as ds
    import datashader.transfer_functions as tf
    from scipy.sparse import issparse
    from bokeh.palettes import RdBu9
    import bokeh
    # 创建一个 Canvas 对象
    cvs = ds.Canvas(plot_width=800, plot_height=800)
    
    
    embedding = adata.obsm[basis]
    # 如果你有一个感兴趣的分类标签，比如细胞类型
    
    # 将数据转换为 DataFrame
    df = pd.DataFrame(embedding, columns=['x', 'y'])

    if color in adata.obs.columns:
        labels = adata.obs[color].tolist()  # 假设'cell_type'是一个列名
    elif color in adata.var_names:
        X=adata[:,color].X
        if issparse(X):
            labels=X.toarray().reshape(-1)
        else:
            labels=X.reshape(-1)
    elif (not adata.raw is None) and (color in adata.raw.var_names):
        X=adata.raw[:,color].X
        if issparse(X):
            labels=X.toarray().reshape(-1)
        else:
            labels=X.reshape(-1)

    
    df['label'] = labels
    #return labels
    #print(labels[0],type(labels[0]))
    if type(labels[0]) is str:
        df['label']=df['label'].astype('category')
        # 聚合数据
        agg = cvs.points(df, 'x', 'y',ds.count_cat('label'),
                        )
        legend_tag=True
        color_key = dict(zip(adata.obs[color].cat.categories,
                        adata.uns[f'{color}_colors']))
        
        
        # 使用色彩映射
        img = tf.shade(tf.spread(agg,px=0),color_key=[color_key[i] for i in df['label'].cat.categories], 
                       how='eq_hist')
    elif (type(labels[0]) is int) or (type(labels[0]) is float) or (type(labels[0]) is np.float32)\
    or (type(labels[0]) is np.float64) or (type(labels[0]) is np.int):
        # 聚合数据
        agg = cvs.points(df, 'x', 'y',ds.mean('label'),
                        )
        legend_tag=False
        if cmap in bokeh.palettes.all_palettes.keys():
            num=list(bokeh.palettes.all_palettes[cmap].keys())[-1]
            img = tf.shade(agg,cmap=bokeh.palettes.all_palettes[cmap][num], 
                           )
        else:
            img = tf.shade(agg,cmap=cmap, 
                           )
    else:
        print('Unrecognized label type')
        return None
    
    
    
        # 假设 img 是 Datashader 渲染的图像
    if ax is None:
        fig, ax = plt.subplots(figsize=figsize)
    else:
        fig=ax.figure
    
    # 假设 img 是一个 NumPy 数组或类似的对象，这里使用 img 的占位符
    # img = np.random.rand(100, 100)  # 示例数据
    ax.imshow(img.to_pil(), aspect='auto')
    
    
    # 自定义格式化函数以显示坐标
    def format_coord(x, y):
        return f"x={x:.2f}, y={y:.2f}"
    
    ax.format_coord = format_coord

    if legend_tag==True:
        # 手动创建图例
        unique_labels = adata.obs[color].cat.categories
        
        # 创建图例项
        for label in unique_labels:
            ax.scatter([], [], c=color_key[label], label=label)
        
        if legend_loc == "right margin":
            ax.legend(
                frameon=False,
                loc="center left",
                bbox_to_anchor=(1, 0.5),
                ncol=(1 if len(unique_labels) <= 14 else 2 if len(unique_labels) <= 30 else 3),
                fontsize=fontsize-1,
            )
    if frameon==False:
        ax.axis('off')
    elif frameon=='small':
        ax.axis('on')
        ax.set_xticks([])
        ax.set_yticks([])
        ax.spines['left'].set_visible(True)
        ax.spines['bottom'].set_visible(True)
        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.spines['bottom'].set_bounds(0,150)
        ax.spines['left'].set_bounds(650,800)
        ax.set_xlabel(f'{basis}1',loc='left',fontsize=fontsize)
        ax.set_ylabel(f'{basis}2',loc='bottom',fontsize=fontsize)

    else:
        ax.axis('on')
        ax.set_xticks([])
        ax.set_yticks([])
        ax.spines['left'].set_visible(True)
        ax.spines['bottom'].set_visible(True)
        ax.spines['top'].set_visible(True)
        ax.spines['right'].set_visible(True)
        ax.set_xlabel(f'{basis}1',loc='center',fontsize=fontsize)
        ax.set_ylabel(f'{basis}2',loc='center',fontsize=fontsize)

    
    # 调整坐标轴线的粗细
    line_width = 1.2  # 设置线宽
    ax.spines['left'].set_linewidth(line_width)
    ax.spines['bottom'].set_linewidth(line_width)
    
    
    if title is None:
        title=color
    ax.set_title(title,fontsize=fontsize+1)

    return ax


r"""
Miscellaneous utilities
"""
import random

from pynvml import *
from anndata import AnnData
from typing import List, Optional, Union

import numpy as np
import torch


def norm_to_raw(
    adata: AnnData, 
    library_size: Optional[Union[str,np.ndarray]] = 'total_counts',
    check_size: Optional[int] = 100
) -> AnnData:
    r"""
    Convert normalized adata.X to raw counts
    
    Parameters
    ----------
    adata
        adata to be convert
    library_size
        raw library size of every cells, can be a key of `adata.obs` or a array
    check_size
        check the head `[0:check_size]` row and column to judge if adata normed
    
    Note
    ----------
    Adata must follow scanpy official norm step 
    """
    check_chunk = adata.X[0:check_size,0:check_size].todense()
    assert not all(isinstance(x, int) for x in check_chunk)
    
    from scipy import sparse
    scale_size = np.array(adata.X.expm1().sum(axis=1).round()).flatten()
    if isinstance(library_size, str):
        scale_factor = np.array(adata.obs[library_size])/scale_size
    elif isinstance(library_size, np.ndarray):
        scale_factor = library_size/scale_size
    else:
        try:
            scale_factor = np.array(library_size)/scale_size
        except:
            raise ValueError('Invalid `library_size`')
    scale_factor.resize((scale_factor.shape[0],1))
    raw_count = sparse.csr_matrix.multiply(sparse.csr_matrix(adata.X).expm1(), sparse.csr_matrix(scale_factor))
    raw_count = sparse.csr_matrix(np.round(raw_count))
    adata.X = raw_count
    # adata.layers['counts'] = raw_count
    return adata


def get_free_gpu() -> int:
    r"""
    Get index of GPU with least memory usage
    
    Ref
    ----------
    https://stackoverflow.com/questions/58216000/get-total-amount-of-free-gpu-memory-and-available-using-pytorch
    """
    nvmlInit()
    index = 0
    max = 0
    for i in range(torch.cuda.device_count()):
        h = nvmlDeviceGetHandleByIndex(i)
        info = nvmlDeviceGetMemoryInfo(h)
        index = i if info.free > max else index
        max = info.free if info.free > max else max
        
    # seed = np.random.randint(1000)
    # os.system(f'nvidia-smi -q -d Memory |grep Used > gpu-{str(seed)}.tmp')
    # memory_available = [int(x.split()[2]) for x in open('gpu.tmp', 'r').readlines()]
    # os.system(f'rm gpu-{str(seed)}.tmp')
    # print(memory_available)
    return index


def global_seed(seed: int) -> None:
    r"""
    Set seed
    
    Parameters
    ----------
    seed 
        int
    """
    if seed > 2**32 - 1 or seed <0:
        seed = 0

    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    if torch.cuda.is_available():
        torch.cuda.manual_seed_all(seed)
    print(f"Global seed set to {seed}.")


import pandas as pd
import numpy as np
import plotly
import os
import matplotlib as mpl
import matplotlib.pyplot as plt
import networkx as nx
from sklearn.neighbors import NearestNeighbors
from sklearn.preprocessing import normalize
from scipy.stats import norm
from networkx.drawing.nx_agraph import to_agraph

def get_cmap_qualitative(cmap_name):
    if cmap_name == "Plotly":
        cmap = plotly.colors.qualitative.Plotly
    elif cmap_name == "Alphabet":
        cmap = plotly.colors.qualitative.Alphabet
    elif cmap_name == "Light24":
        cmap = plotly.colors.qualitative.Light24
    elif cmap_name == "Dark24":
        cmap = plotly.colors.qualitative.Dark24
    # Safe and Vivid are strings of form "rbg(...)"
    # Handle this later.
    elif cmap_name == "Safe":
        cmap = plotly.colors.qualitative.Safe
    elif cmap_name == "Vivid":
        cmap = plotly.colors.qualitative.Vivid
    return cmap

def linear_clamp_value(x, lower_bound, upper_bound, out_min, out_max):
    if x <= lower_bound:
        y = out_min
    elif x >= upper_bound:
        y = out_max
    else:
        y = out_min + (x - lower_bound)/(upper_bound-lower_bound) * (out_max-out_min)
    return y


def plot_cluster_signaling_chord(S, p_values, label_name=None, colormap='Plotly', quantile_cutoff=None, p_value_cutoff=None, cutoff=0, separate=False, filename="chord_plot.pdf", diagonal_off=False):
    cmap = get_cmap_qualitative(colormap)
    data_all = np.empty([0,3], str)
    if not quantile_cutoff is None:
        cutoff = np.quantile(S.reshape(-1), quantile_cutoff)
    for i in range(S.shape[0]):
        for j in range(S.shape[1]):
            if diagonal_off and i==j: continue
            if S[i,j] >= cutoff or p_values[i,j] <= p_value_cutoff:
                if not label_name is None:
                    if separate:
                        data = np.array([['S'+label_name[i], 'T'+label_name[j], str(S[i,j])]], str)
                    else:
                        data = np.array([[label_name[i], label_name[j], str(S[i,j])]], str)
                else:
                    if separate:
                        data = np.array([['S'+str(i), 'T'+str(j), str(S[i,j])]], str)
                    else:
                        data = np.array([[str(i), str(j), str(S[i,j])]], str)
                data_all = np.concatenate((data_all, data), axis=0)
    df = pd.DataFrame(data=data_all, columns=['from','to','value'])
    df.to_csv("./plot_chord_data.csv", index=False)

    data_cmap = np.empty([0, 2], str)
    for i in range(S.shape[0]):
        if not label_name is None:
            if separate:
                data_tmp = np.array([['S'+label_name[i], cmap[i % len(cmap)]],['T'+label_name[i], cmap[i % len(cmap)]]], str)
            else:
                data_tmp = np.array([[label_name[i], cmap[i % len(cmap)]]], str)
        else:
            if separate:
                data_tmp = np.array([['S'+str(i), cmap[i % len(cmap)]],['T'+str(i), cmap[i % len(cmap)]]], str)
            else:
                data_tmp = np.array([[str(i), cmap[i % len(cmap)]]], str)
        data_cmap = np.concatenate((data_cmap, data_tmp), axis=0)
    df = pd.DataFrame(data=data_cmap)
    df.to_csv("./plot_chord_cmap.csv", index=False, header=False)

    os.system("Rscript plot_chord.R")
    os.system("mv Rplots.pdf "+filename)
    os.system("pdfcrop %s %s" % (filename, filename))
    filename_png = filename[:-4]
    filename_png = filename_png + ".png"
    os.system("convert -density 800 %s -quality 100 %s" % (filename, filename_png))

def plot_cluster_signaling_network(S,
    labels = None,
    node_size = 0.2,
    node_colormap = "Plotly",
    node_cluster_colormap = None,
    node_pos = None,
    edge_width_lb_quantile = 0.05,
    edge_width_ub_quantile = 0.95,
    edge_width_min = 1,
    edge_width_max = 4,
    edge_color = None, # expect to range from 0 to 1
    edge_colormap = None,
    background_pos = None,
    background_ndcolor = "lavender",
    background_ndsize = 1,
    filename = "network_plot.pdf",
):
    if labels is None:
        labels = [str(i) for i in range(S.shape[0])]
    node_cmap = get_cmap_qualitative(node_colormap)
    G = nx.MultiDiGraph()

    edge_width_lb = np.quantile(S.reshape(-1), edge_width_lb_quantile)
    edge_width_ub = np.quantile(S.reshape(-1), edge_width_ub_quantile)


    # Draw the background geometry
    if not background_pos is None:
        for i in range(background_pos.shape[0]):
            G.add_node("cell_"+str(i), shape='point', color=background_ndcolor, fillcolor=background_ndcolor, width=background_ndsize)
            G.nodes["cell_"+str(i)]["pos"] = "%f,%f!" %(background_pos[i,0],background_pos[i,1])

    # Draw the nodes (cluster)
    for i in range(len(labels)):
        if node_cluster_colormap is None:
            G.add_node(labels[i], shape="point", fillcolor=node_cmap[i], color=node_cmap[i])
        elif not node_cluster_colormap is None:
            G.add_node(labels[i], shape="point", fillcolor=node_cluster_colormap[labels[i]], color=node_cmap[i])
        if not node_pos is None:
            G.nodes[labels[i]]["pos"] = "%f,%f!" % (node_pos[i,0],node_pos[i,1])
        G.nodes[labels[i]]["width"] = str(node_size)

    # Draw the edges
    for i in range(S.shape[0]):
        for j in range(S.shape[1]):
            if S[i,j] > 0:
                G.add_edge(labels[i], labels[j], splines="curved")
                G[labels[i]][labels[j]][0]["penwidth"] = str(linear_clamp_value(S[i,j],edge_width_lb,edge_width_ub,edge_width_min,edge_width_max))
                if edge_color == "node":
                    G[labels[i]][labels[j]][0]['color'] = node_cmap[i]
                elif isinstance(edge_color, np.ndarray):
                    G[labels[i]][labels[j]][0]['color'] = mpl.colors.to_hex( edge_colormap(edge_color[i,j]) )
                else:
                    G[labels[i]][labels[j]][0]['color'] = edge_color
    
    # Draw the network
    A = to_agraph(G)
    if node_pos is None:
        A.layout("dot")
    else:
        A.layout()
    A.draw(filename)

# Recommend plotting at most three LR species
def plot_cluster_signaling_network_multipair(S_list,
    p_values_list,
    labels = None,
    quantile_cutoff = None,
    p_value_cutoff = 0.05,
    node_size = 0.2,
    node_colormap = "Plotly",
    node_pos = None,
    edge_width_lb_quantile = 0.05,
    edge_width_ub_quantile = 0.95,
    edge_width_min = 1,
    edge_width_max = 4,
    edge_colors = ["blue","red","green","grey","yellow"],
    background_pos = None,
    background_ndcolor = "darkseagreen1",
    background_ndsize = 1,
    filename = "network_plot.pdf"
):
    tmp_all_S = []
    for S in S_list:
        tmp_all_S.extend( list( S.reshape(-1) ) )
    tmp_all_S = np.array(tmp_all_S)
    if labels is None:
        labels = [str(i) for i in range(S.shape[0])]
    node_cmap = get_cmap_qualitative(node_colormap)
    G = nx.MultiDiGraph()
    if not quantile_cutoff is None:
        cutoff = np.quantile(tmp_all_S.reshape(-1), quantile_cutoff)
    else:
        cutoff = 0.0
    edge_width_lb = np.quantile(tmp_all_S.reshape(-1), edge_width_lb_quantile)
    edge_width_ub = np.quantile(tmp_all_S.reshape(-1), edge_width_ub_quantile)

    # Draw the background geometry
    if not background_pos is None:
        for i in range(background_pos.shape[0]):
            G.add_node("cell_"+str(i), shape='point', color=background_ndcolor, fillcolor=background_ndcolor, width=background_ndsize)
            G.nodes["cell_"+str(i)]["pos"] = "%f,%f!" %(background_pos[i,0],background_pos[i,1])

    # Draw the nodes (cluster)
    for i in range(len(labels)):
        G.add_node(labels[i], shape="point", fillcolor=node_cmap[i], color=node_cmap[i])
        if not node_pos is None:
            G.nodes[labels[i]]["pos"] = "%f,%f!" % (node_pos[i,0],node_pos[i,1])
        G.nodes[labels[i]]["width"] = str(node_size)

    # Draw the edges
    for k in range(len(S_list)):
        S = S_list[k]
        p_values = p_values_list[k]
        for i in range(S.shape[0]):
            for j in range(S.shape[1]):
                if S[i,j] >= cutoff or p_values[i,j] <= p_value_cutoff:
                    penwidth = str(linear_clamp_value(S[i,j],edge_width_lb,edge_width_ub,edge_width_min,edge_width_max))
                    color = edge_colors[k]
                    G.add_edge(labels[i], labels[j], splines="curved", penwidth=penwidth, color=color)
    
    # Draw the network
    A = to_agraph(G)
    if node_pos is None:
        A.layout("dot")
    else:
        A.layout()
    A.draw(filename)

def plot_cell_signaling(X,
    V,
    signal_sum,
    cmap="coolwarm",
    cluster_cmap = None,
    arrow_color="tab:blue",
    plot_method="cell",
    background='summary',
    clustering=None,
    background_legend=False,
    adata=None,
    summary='sender',
    ndsize = 1,
    scale = 1.0,
    grid_density = 1,
    grid_knn = None,
    grid_scale = 1.0,
    grid_thresh = 1.0,
    grid_width = 0.005,
    stream_density = 1.0,
    stream_linewidth = 1,
    stream_cutoff_perc = 5,
    filename=None,
    ax = None,
    fig = None
):
    ndcolor = signal_sum
    ncell = X.shape[0]
    
    V_cell = V.copy()
    V_cell_sum = np.sum(V_cell, axis=1)
    V_cell[np.where(V_cell_sum==0)[0],:] = np.nan
    if summary == "sender":
        X_vec = X
    elif summary == "receiver":
        X_vec = X - V / scale

    if plot_method == "grid" or plot_method == "stream":
        # Get a rectangular grid
        xl, xr = np.min(X[:,0]), np.max(X[:,0])
        epsilon = 0.02*(xr-xl); xl -= epsilon; xr += epsilon
        yl, yr = np.min(X[:,1]), np.max(X[:,1])
        epsilon = 0.02*(yr-yl); yl -= epsilon; yr += epsilon
        ngrid_x = int(50 * grid_density)
        gridsize = (xr-xl) / float(ngrid_x)
        ngrid_y = int((yr-yl)/gridsize)
        meshgrid = np.meshgrid(np.linspace(xl,xr,ngrid_x), np.linspace(yl,yr,ngrid_y))
        grid_pts = np.concatenate((meshgrid[0].reshape(-1,1), meshgrid[1].reshape(-1,1)), axis=1)
    
        if grid_knn is None:
            grid_knn = int( X.shape[0] / 50 )
        nn_mdl = NearestNeighbors()
        nn_mdl.fit(X)
        dis, nbs = nn_mdl.kneighbors(grid_pts, n_neighbors=grid_knn)
        w = norm.pdf(x=dis, scale=gridsize * grid_scale)
        w_sum = w.sum(axis=1)

        V_grid = (V[nbs] * w[:,:,None]).sum(axis=1)
        V_grid /= np.maximum(1, w_sum)[:,None]

        if plot_method == "grid":
            grid_thresh *= np.percentile(w_sum, 99) / 100
            grid_pts, V_grid = grid_pts[w_sum > grid_thresh], V_grid[w_sum > grid_thresh]
        elif plot_method == "stream":
            x_grid = np.linspace(xl, xr, ngrid_x)
            y_grid = np.linspace(yl, yr, ngrid_y)
            V_grid = V_grid.T.reshape(2, ngrid_y, ngrid_x)
            vlen = np.sqrt((V_grid ** 2).sum(0))
            grid_thresh = 10 ** (grid_thresh - 6)
            grid_thresh = np.clip(grid_thresh, None, np.max(vlen) * 0.9)
            cutoff = vlen.reshape(V_grid[0].shape) < grid_thresh
            length = np.sum(np.mean(np.abs(V[nbs]),axis=1),axis=1).T
            length = length.reshape(ngrid_y, ngrid_x)
            cutoff |= length < np.percentile(length, stream_cutoff_perc)

            V_grid[0][cutoff] = np.nan

    if cmap == 'Plotly':
        cmap = plotly.colors.qualitative.Plotly
    elif cmap == 'Light24':
        cmap = plotly.colors.qualitative.Light24
    elif cmap == 'Dark24':
        cmap = plotly.colors.qualitative.Dark24
    elif cmap == 'Alphabet':
        cmap = plotly.colors.qualitative.Alphabet

    idx = np.argsort(ndcolor)
    if background == 'summary' or background == 'cluster':
        if not ndsize==0:
            if background == 'summary':
                ax.scatter(X[idx,0], X[idx,1], s=ndsize, c=ndcolor[idx], cmap=cmap, linewidth=0)
            elif background == 'cluster':
                adata.obs[clustering]=adata.obs[clustering].astype('category')
                labels = np.array( adata.obs[clustering], str )
                unique_labels = adata.obs[clustering].cat.categories
                for i_label in range(len(unique_labels)):
                    idx = np.where(labels == unique_labels[i_label])[0]
                    if cluster_cmap is None:
                        ax.scatter(X[idx,0], X[idx,1], s=ndsize, c=cmap[i_label], linewidth=0, label=unique_labels[i_label])
                    elif not cluster_cmap is None:
                        ax.scatter(X[idx,0], X[idx,1], s=ndsize, c=cluster_cmap[unique_labels[i_label]], linewidth=0, label=unique_labels[i_label])
                if background_legend:
                    ax.legend(markerscale=2.0, loc=[1.0,0.0])
        if plot_method == "cell":
            ax.quiver(X_vec[:,0], X_vec[:,1], V_cell[:,0], V_cell[:,1], scale=scale, scale_units='x', color=arrow_color)
        elif plot_method == "grid":
            ax.quiver(grid_pts[:,0], grid_pts[:,1], V_grid[:,0], V_grid[:,1], scale=scale, scale_units='x', width=grid_width, color=arrow_color)
        elif plot_method == "stream":
            lengths = np.sqrt((V_grid ** 2).sum(0))
            stream_linewidth *= 2 * lengths / lengths[~np.isnan(lengths)].max()
            ax.streamplot(x_grid, y_grid, V_grid[0], V_grid[1], color=arrow_color, density=stream_density, linewidth=stream_linewidth)
    elif background == 'image':
        spatial_mapping = adata.uns.get("spatial", {})
        library_id = list(spatial_mapping.keys())[0]
        spatial_data = spatial_mapping[library_id]
        img = spatial_data['images']['hires']
        sf = spatial_data['scalefactors']['tissue_hires_scalef']
        ax.imshow(img, origin='lower')
        if plot_method == "cell":
            ax.quiver(X_vec[:,0]*sf, X_vec[:,1]*sf, V_cell[:,0]*sf, V_cell[:,1]*sf, scale=scale, scale_units='x', color=arrow_color)
        elif plot_method == "grid":
            ax.quiver(grid_pts[:,0]*sf, grid_pts[:,1]*sf, V_grid[:,0]*sf, V_grid[:,1]*sf, scale=scale, scale_units='x', width=grid_width, color=arrow_color)
        elif plot_method == "stream":
            lengths = np.sqrt((V_grid ** 2).sum(0))
            stream_linewidth *= 2 * lengths / lengths[~np.isnan(lengths)].max()
            ax.streamplot(x_grid*sf, y_grid*sf, V_grid[0]*sf, V_grid[1]*sf, color=arrow_color, density=stream_density, linewidth=stream_linewidth)

    ax.axis("equal")
    ax.axis("off")
    if not filename is None:
        plt.savefig(filename, dpi=500, bbox_inches = 'tight', transparent=True)


def plot_cell_signaling_compare(X,
    V,
    ax,
    arrow_color="tab:blue",
    plot_method="cell",
    summary='sender',
    scale = 1.0,
    grid_density = 1,
    grid_knn = None,
    grid_scale = 1.0,
    grid_thresh = 1.0,
):
    V_cell = V.copy()
    V_cell_sum = np.sum(V_cell, axis=1)
    V_cell[np.where(V_cell_sum==0)[0],:] = np.nan
    if summary == "sender":
        X_vec = X
    elif summary == "receiver":
        X_vec = X - V / scale

    if plot_method == "grid":
        # Get a rectangular grid
        xl, xr = np.min(X[:,0]), np.max(X[:,0])
        epsilon = 0.02*(xr-xl); xl -= epsilon; xr += epsilon
        yl, yr = np.min(X[:,1]), np.max(X[:,1])
        epsilon = 0.02*(yr-yl); yl -= epsilon; yr += epsilon
        ngrid_x = int(50 * grid_density)
        gridsize = (xr-xl) / float(ngrid_x)
        ngrid_y = int((yr-yl)/gridsize)
        meshgrid = np.meshgrid(np.linspace(xl,xr,ngrid_x), np.linspace(yl,yr,ngrid_y))
        grid_pts = np.concatenate((meshgrid[0].reshape(-1,1), meshgrid[1].reshape(-1,1)), axis=1)
    
        if grid_knn is None:
            grid_knn = int( X.shape[0] / 50 )
        nn_mdl = NearestNeighbors()
        nn_mdl.fit(X)
        dis, nbs = nn_mdl.kneighbors(grid_pts, n_neighbors=grid_knn)
        w = norm.pdf(x=dis, scale=gridsize * grid_scale)
        w_sum = w.sum(axis=1)

        V_grid = (V[nbs] * w[:,:,None]).sum(axis=1)
        V_grid /= np.maximum(1, w_sum)[:,None]

        if plot_method == "grid":
            grid_thresh *= np.percentile(w_sum, 99) / 100
            grid_pts, V_grid = grid_pts[w_sum > grid_thresh], V_grid[w_sum > grid_thresh]

    if plot_method == "cell":
        ax.quiver(X_vec[:,0], X_vec[:,1], V_cell[:,0], V_cell[:,1], scale=scale, scale_units='x', color=arrow_color, alpha=1.0)
    elif plot_method == "grid":
        ax.quiver(grid_pts[:,0], grid_pts[:,1], V_grid[:,0], V_grid[:,1], scale=scale, scale_units='x', color=arrow_color, alpha=1.0)
    return ax